// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

/*! \file lp.h
 * \brief Header for LP class.
 */

#ifndef _LP_H_
#define _LP_H_

//STL includes
#include<string>
#include<list>
#include<ostream>
#include<map>
#include<limits>

/// Definition of ROPI positive infinity
#define ROPIPINF std::numeric_limits<double>::max( )

/// Definition of ROPI negative infinity
#define ROPIMINF std::numeric_limits<double>::min( )

namespace ROPI
{
  class LPimpl;
  
  /// Variable type enumerator. 
  enum VAR_TYPE
  {
    /// Continuous variable.
    CONTINUOUS = 1,
    
    /// Integer variable.
    INTEGER = 2,
    
    /// Binary variable.
    BINARY = 3
  };
  
  /// Type of program enumerator.
  enum LPTYPE
  {
    /// Linear Program.
    LPTYPE_LP = 1,
    
    /// (Mixed) Integer Program.
    LPTYPE_MIP = 2
  };

  /// Variable data structure
  struct Var
  {
	/// Relational operator.
    bool operator< (const Var& other) const
    {
      return (id < other.id);
    }
    
    /// Equality check.
    bool operator== (const Var& other) const
    {
      return (id == other.id);
    }
    
    /// Default constructor.
    Var() : obj(0), id(-1), name("DEFAULTNAME"), lb(0), ub(ROPIPINF), type(CONTINUOUS), solution(0), start_given(false), start(0) {};
    
    /// \brief Objective coefficient.
    /// Default is 0.
    double obj;
    
    /// \brief Unique identificator number.
    /// ids must start from 0 and
    /// must be numbered consecutively.
    /// Default is -1.
    int id;
    
    /// \brief Variable name.
    /// Default is "DEFAULTNAME".
    std::string name;
    
    /// \brief Variable lower bound.
    /// Default is 0.
    double lb;
    
    /// \brief Variable upper bound.
    /// Default is ROPIPINF.
    double ub;
    
    /// \brief Variable type (continuous, integer, binary).
    /// Default is continuous.
    VAR_TYPE type;
    
    /// \brief The value of the variable in the current solution.
    /// Default is 0.
    double solution;
    
    /// \brief Flag if a start solution provided.
    /// Default is 0.
    bool start_given;
    
    /// \brief Variable value in start solution.
    /// Default is 0.
    double start;
  };

  /// Constraint sense enumerator.
  enum CON_SENSE
  {
    /// Less or equal constraint.
    LEQ = 1,
    
    /// Equality constraint.
    EQ = 2,
    
    /// Greater or equal constraint.
    GEQ = 3
  };

  /// Objective direction.
  enum OBJ_SENSE
  {
    /// Maximization problem
    MAX = 1,
    
    /// Minimization problem
    MIN = 2
  };

  /// Constraint data structure.
  struct Con
  {
	/// Relational operator.
    bool operator< (const Con& other) const
    {
      return (id < other.id);
    }
    
    /// Default constructor.
    Con() : rhs(0), sen(EQ), id(-1), name("DEFAULTNAME") {};
    
    /// \brief Right-hand-side of constraint.
    /// Default is 0.
    double rhs;
    
    /// \brief Constraint sense (<=, =, >=).
    /// Default is equality.
    CON_SENSE sen;
    
    /// \brief Unique identificator number.
    /// ids must start from 0 and
    /// must be numbered consecutively.
    /// Default is -1.
    int id;
    
    /// Constraint name. Default is "DEFAULTNAME".
    std::string name;
    
    /// \brief Variable coefficients of constraint.
    /// The pairs consist of variable id and coefficient.
    std::list<std::pair<int,double>  > coeffs;
  };

  /// Class for (mixed integer) linear programs.
  class LP
  {
    public:
      /// Default constructor.
      LP();
      
      /// Default destuctor.
      ~LP();
      
      /// \brief Adds a variable to the problem.
      /// Note: a copy of the variable will be created.
      /// Ownership of the argument is preserved.
      void add_var(Var _var);
      
      /// \brief Adds a constraint to the problem.
      /// Note: a copy of the constraint will be created.
      /// Ownership of the argument is preserved.
      void add_con(Con _con);
      
      /// Set the problem name. Default is "DEFAULTNAME".
      void set_name(std::string _name);
      
      /// Set the problem direction (MIN,MAX). Default is MIN.
      void set_sense(OBJ_SENSE _sen);
      
      /// Returns the problem name.
      std::string get_name();
      
      /// Returns the problem direction.
      OBJ_SENSE get_sense();
      
      /// Returns a pointer to the list of variables.
      std::list<Var>* get_vars();
      
      /// Returns a pointer to the list of constraints.
      std::list<Con>* get_cons();
      
      /// Returns the number of variables.
      int get_num_vars();
      
      /// Returns the number of constraints.
      int get_num_cons();
      
      /// Returns the type of the problem (LP,MIP).
      LPTYPE get_lptype();
      
      /// Read a problem from an mps file (fixed or free).
      void read_mpsfile(std::string file);
      
      /// Write a human-friendly version of the problem to ostream (e.g., cout).
      void write_plain(std::ostream &out);
      
    private:
      /// Pointer to implementation details.
      LPimpl* lpimpl;
  };
}
#endif
