// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

/*! \file robustness.h
 * \brief Header for robustness class.
 */

#ifndef _H_ROBUSTNESS_H_
#define _H_ROBUSTNESS_H_

#include "lp.h"
#include "solver.h"
#include<vector>
#include<map>

namespace ROPI
{  
  class RobustImpl;
  
  /// Robustness model enumerator.
  enum Robustness_Method
  {
	/// Strict robustness.
    ROB_STRICT = 1,
    
    /// Light robustness.
    ROB_LIGHT = 2,
    
    /// Recovery to optimality.
    ROB_RECOPT = 3,
    
    /// Recovery to feasibility.
    ROB_RECFEAS = 4,
    
    /// Undefined robustness model.
    ROB_UNDEFINED = 5
  };
  
  /// Uncertainty type enumerator.
  enum Uncertainty_Type
  {
	/// Finite uncertainty.
    UNC_FINITE = 1,
    
    /// Undefined uncertainty type.
    UNC_UNDEFINED = 2
  };
  
  /// Norm enumerator.
  enum Norm
  {
	/// l_1 norm.
    NORM_L1 = 1,
    
    /// l_infty norm.
    NORM_LINFTY = 2,
    
    /// Undefined norm.
    NORM_UNDEFINED = 3
  };
  
  /// Recovery objective enumerator.
  enum RecoveryObjective
  {
	/// Median objective.
    REC_MEDIAN = 1,
    
    /// Center objective.
    REC_CENTER = 2,
    
    /// Undefined objective.
    REC_UNDEFINED = 3
  };
  
  /// Recovery to optimality model enumerator.
  enum Recopt_Model
  {
	  /// Simplex recovery to optimality.
	  RECOPT_SIMPLE = 1,
	  
	  /// Extended recovery to optimality.
	  RECOPT_EXTENDED = 2,
	  
	  /// Undefined RecOpt model.
	  RECOPT_UNDEFINED = 3
  };
  
  /// A single scenario.
  struct FiniteScenario
  {
	/// \brief Map containing all coefficients that differ from the nominal case.
	/// Con is the affected constraint, int the variable id.
    std::map<std::pair<Con,int>, double> coeffs;
    
    /// Map containing all right-hand-sides that differ from the nominal case.
    std::map<Con,double> rhs;
  };
  
  /// A finite uncertainty set.
  struct FiniteUncertainty
  {
	/// Finite uncertainty consists of a vector of scenarios.
    std::vector<FiniteScenario> scenarios;
  };
  
  /// Options for creating the robust counterpart.
  struct RobustnessOptions
  {
	/// Default constructor.
    RobustnessOptions() : 
		norm(NORM_UNDEFINED),
		recobj(REC_UNDEFINED),
		solvertype(SOL_UNDEFINED),
		recopt_model(RECOPT_UNDEFINED),
		nominal_feasibility(true),
		budget(1.5) {};
    
    /// \brief Recovery norm. Default is NORM_UNDEFINED.
    /// Relevant for RecOpt and RecFeas.
    Norm norm;
    
    /// \brief Recovery objective. Default is REC_UNDEFINED.
    /// Relevant for RecOpt and RecFeas.
    RecoveryObjective recobj;
    
    /// \brief The solver used for preprocessing. Default is SOL_UNDEFINED.
    /// Relevant for Light Robustness and RecOpt.
    SOLVERTYPE solvertype;   
    
    /// The RecOpt model applied. Default is RECOPT_UNDEFINED.
    Recopt_Model recopt_model; 
    
    /// \brief Flag that determines if recovery robust solutions
    /// need to be feasible for the nominal scenario.
    /// Default is true.
    /// Relevant for RecOpt and RecFeas.
    bool nominal_feasibility;
    
    /// \brief Determines the budget for light robustness
    /// (budget) * (nominal objective value) is the worst objective for light robust solution.
    /// Default is 1.5.
    double budget;
    
  };
  
  /// Robustness class.
  class Robustness
  {
    public:
      /// Default constructor. LP ownership is preserved.
      Robustness(LP* _lp, Robustness_Method method);
      
      /// Set the uncertainty.
      void set_uncertainty(FiniteUncertainty _unc);
      
      /// Set robustness options.
      void set_options(RobustnessOptions _options);
     
      /// Generate the robust counterpart.
      LP generate_robust();
      
      /// RecFeas specific: Exclude variable from objective function.
      void exclude_variable(int id);
      
    private:
      RobustImpl* rimpl;
  };
  
}


#endif
