// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

/*! \file solver.h
 * \brief Header for solver class.
 */

#ifndef _SOLVER_H_
#define _SOLVER_H_

//STL includes
#include<list>
#include<ostream>
#include<vector>

namespace ROPI
{
  class LP;
  
  class SolverImpl;
  
  /// Type of solver enumerator.
  enum SOLVERTYPE
  {
    /// Gurobi
    SOL_GUROBI = 1,
    
    /// Cplex
    SOL_CPLEX = 2,
    
    /// Xpress
    SOL_XPRESS = 3,
    
    /// Undefined
    SOL_UNDEFINED = 4
  };
  
  /// Solver status enumerator.
  enum SOLVERSTATUS
  {
    /// Empty: No problem loaded.
    SOL_EMPTY = 1,
    
    /// Loaded: Problem loaded.
    SOL_LOADED = 2,
    
    /// Optimal: Problem solved to optimality.
    SOL_OPTIMAL = 3,
    
    /// Infeasible: Problem is not feasible.
    SOL_INFEASIBLE = 4,
    
    /// Unbounded: Problem is unbounded.
    SOL_UNBOUNDED = 5,
    
    /// Timelimit: Solution process was aborted due to time constraint.
    SOL_TIMELIMIT = 6,
    
    /// Unknown: Status is unknown
    SOL_UNKNOWN = 7
  };
  
  /// Solver options data structure.
  struct SolverOptions
  {
    /// Default constructor.
    SolverOptions() : verbose(true), timelimit(0), threads(0), mipfocus(0), heuristics(-1) {};
    
    /// Flag to turn solver output on/off. Default is true.
    bool verbose;
    
    /// Solver timelimit. 0 means no limit. Default is 0.
    int timelimit;
    
    /// Number of threads the solver may use. 0 means automatic. Default is 0.
    int threads;
    
    /// \brief Solver focus.
    /// 0 means automatic.
    /// 1 means improve current solution.
    /// 2 means prove optimality.
    /// 3 means improve lower bound.
    /// Default is 0.
    /// Warning: Currently only available for Gurobi.
    int mipfocus;
    
    /// \brief Heuristics parameter.
    /// Determines the percentage of time for heuristics
    /// -1 uses the solver presetting.
    /// Default is -1.
    /// Warning: Currently only available for Gurobi.
    double heuristics;
  };
  
  /// Solver class
  class Solver
  {
    public:
      /// Default constructor.
      Solver();
      
      /// Default destructor.
      ~Solver();
      
      /// \brief Initialize the solver-specific program.
      /// LP ownership is preserved.
      void init(LP* _lp, SOLVERTYPE _type);
      
      /// Solve the problem.
      void solve();
      
      /// \brief Write solution in human-friendly format to stream.
      /// Solution values with |x| <= eps are not shown.
      void write_solution(std::ostream &out, double eps);
      
      /// Set solver options.
      void set_options(SolverOptions _options);
      
      /// \brief Returns a list of currently available solvers.
      /// Note that solver availability is determined during compilation .
      static std::list<SOLVERTYPE> get_available();
      
      /// Returns the current solver status.
      SOLVERSTATUS get_status();
      
      /// Returns the objective value of the current solution.
      double get_objective();
      
      /// Returns the current solution.
      std::vector<double> get_solution();
      
    private:
      SOLVERTYPE type;
      
      SolverImpl* solver_impl;
  };
}

#endif
