// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "lpimpl.h"
#include "error.h"
#include<fstream>
#include<iostream>
#include<sstream>
#include<cstdlib>
#include<cmath>

using namespace std;
using namespace ROPI;

void LPimpl::add_var(Var _var)
{
  vars.push_back(_var);
  ++num_vars;
  
  if (_var.obj != 0)
    obj.push_back(&(vars.back()));
  
  idvarmap[_var.id] = &vars.back();
  
  if (_var.type == BINARY || _var.type == INTEGER)
    lptype = LPTYPE_MIP;
  
}

list<Var>* LPimpl::get_vars()
{
  return &vars;
}

int LPimpl::get_num_vars()
{
  return num_vars;
}

void LPimpl::add_con(Con _con)
{
  cons.push_back(_con);
  ++num_cons;
}

list<Con>* LPimpl::get_cons()
{
  return &cons;
}

int LPimpl::get_num_cons()
{
  return num_cons;
}

void LPimpl::set_sense(OBJ_SENSE _sen)
{
  obj_sen = _sen;
}

OBJ_SENSE LPimpl::get_sense()
{
  return obj_sen;
}

void LPimpl::set_name(string _name)
{
  name = _name;
}

string LPimpl::get_name()
{
  return name;
}

LPTYPE LPimpl::get_lptype()
{
  return lptype;
}

void LPimpl::read_mpsfile(string file)
{
  ifstream in(file.c_str());
  
  MPS_READMODE readmode = MPS_NONE;
  string line;
  bool fixed = true;

  while (fixed && !in.eof())
  {
    getline(in,line);
    
    //empty lines
    if (line.size() == 0)
      continue;
    
    //comments
    if ((line.c_str())[0]=='*')
      continue;
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTORG'")!=string::npos)
      continue;
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTEND'")!=string::npos)
      continue;
    
    //header
    if ((line.c_str())[0]!=' ')
    {
      if (line.substr(0,4) == "NAME")
      {
	if (line.substr(4,10) != "          ")
	  fixed = false;
      }
      else if (line.substr(0,4) == "ROWS")
	readmode = MPS_ROWS;
      else if (line.substr(0,7) == "COLUMNS")
	readmode = MPS_COLUMNS;
      else if (line.substr(0,3) == "RHS")
	readmode = MPS_RHS;
      else if (line.substr(0,6) == "RANGES")
	readmode = MPS_RANGES;
      else if (line.substr(0,6) == "BOUNDS")
	readmode = MPS_BOUNDS;
      else if (line.substr(0,6) == "ENDATA")
	readmode = MPS_NONE;
      else
      {
	  fixed = false;
      }
      
      if (!fixed)
	cout<<"Unfixed in header.\n";
      
    }
    else //no header
    {
      switch (readmode)
      {
	case MPS_ROWS:
	{
	  if (line.size() > 12)
	    fixed = false;
	}
	if (!fixed)
	cout<<"Unfixed in rows.\n";
	
	  break;

	case MPS_COLUMNS:
	{
	  if (line.substr(0,4) != "    ")
	    fixed = false;
	  
	  if (line.substr(12,2) != "  ")
	    fixed = false;

	  if (line.substr(22,2) != "  ")
	    fixed = false;

	  if (line.size() > 46)
	  {
	    if (line.substr(47,2) != "  ")
	      fixed = false;
	  }
	}
	if (!fixed)
	  cout<<"Unfixed in columns.\n";
	  break;

	case MPS_RHS:
	{
	  if (line.substr(0,4) != "    ")
	    fixed = false;
	  
	  if (line.substr(12,2) != "  ")
	    fixed = false;

	  if (line.substr(22,2) != "  ")
	    fixed = false;

	  if (line.size() > 46)
	  {
	    if (line.substr(47,2) != "  ")
	      fixed = false;
	  }
	}
	if (!fixed)
	  cout<<"Unfixed in rhs.\n";
	  break;

	case MPS_RANGES:
	{
	  if (line.substr(0,4) != "    ")
	    fixed = false;
	  
	  if (line.substr(12,2) != "  ")
	    fixed = false;

	  if (line.substr(22,2) != "  ")
	    fixed = false;

	  if (line.size() > 46)
	  {
	    if (line.substr(47,2) != "  ")
	      fixed = false;
	  }
	}
	if (!fixed)
	  cout<<"Unfixed in ranges.\n";
	  break;

	case MPS_BOUNDS:
	{
	  if (line.substr(0,1) != " ")
	    fixed = false;
	  
	  if (line.substr(3,1) != " ")
	    fixed = false;

	  if (line.substr(12,2) != "  ")
	    fixed = false;

	  if (line.size() > 22)
	  {
	    if (line.substr(22,2) != "  ")
	      fixed = false;
	    
	    if (line.size() > 36)
		fixed = false;
	  }
	}
	if (!fixed)
	  cout<<"Unfixed in bounds.\n";
	  break;

	default:
	{
	  fixed = false;
	  if (!fixed)
	    cout<<"Unfixed in default.\n";
	}
      }
    }
  }

  
  in.close();
  
  if (fixed)
    read_mpsfile_fixed(file);
  else
    read_mpsfile_free(file);
  
}

void LPimpl::read_mpsfile_fixed(string file)
{
//   cout<<"Reading in file "<<file<<"\n";
  cout<<"Using fixed mps.\n";
  
  ifstream in(file.c_str());
  
  MPS_READMODE readmode = MPS_NONE;
  string line;
  int linenumber = 0;
  int varid = 0;
  int conid = 0;
  string obj_name;
  map<string,Var> namevarmap;
  map<string,Con> nameconmap;
  obj_sen = MIN;
  bool intblock = false;
  
  while (!in.eof())
  {
    getline(in,line);
    ++linenumber;
//     cout<<linenumber<<"\n"<<flush;
    
    vector<string> tokens = tokenize(line);
    
    //empty lines
    if (tokens.size() == 0)
      continue;
    
    //comments
    if ((line.c_str())[0]=='*')
      continue;
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTORG'")!=string::npos)
    {
      intblock = true;
      continue;
    }
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTEND'")!=string::npos)
   {
      intblock = false;
      continue;
    }
    
    //header
    if ((line.c_str())[0]!=' ')
    {
      if (tokens[0] == "NAME")
      {
	if (line.size() < 15)
	{
	  stringstream estring;
	  estring << "In read_mpsfile: NAME is not followed by string.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
	}
	name = line.substr(14,8);
	readmode = MPS_NAME;
      }
      else if (tokens[0] == "ROWS")
	readmode = MPS_ROWS;
      else if (tokens[0] == "COLUMNS")
	readmode = MPS_COLUMNS;
      else if (tokens[0] == "RHS")
	readmode = MPS_RHS;
      else if (tokens[0] == "RANGES")
	readmode = MPS_RANGES;
      else if (tokens[0] == "BOUNDS")
	readmode = MPS_BOUNDS;
      else if (tokens[0] == "ENDATA")
	readmode = MPS_NONE;
      else
      {
	  stringstream estring;
	  estring << "In read_mpsfile: Unknown header.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
      }
    }
    else //no header
    {
      switch (readmode)
      {
	case MPS_NAME:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Name is not part of name header.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
	  break;
	}
	case MPS_ROWS:
	{
	  string conname = line.substr(4,8);
	  if (conname.length() < 8)
	    conname.append(8-conname.length(),' ');

	  switch ((line.c_str())[1])
	  {
	    case 'N':
	      obj_name = conname;
// 	      cout<<"Name of Objective: "<<obj_name<<"\n";
	      break;
	    case 'L':
	    {
	      Con con;
	      con.sen = LEQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    case 'G':
	    {
	      Con con;
	      con.sen = GEQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    case 'E':
	    {
	      Con con;
	      con.sen = EQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    default:
	    {
	      stringstream estring;
	      estring << "In read_mpsfile: Unknown row type.\n";
	      estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	      throw Error(estring.str());
	    }
	  }
	}
	  break;
	case MPS_COLUMNS:
	{
	  string var_name = line.substr(4,8);
// 	  cout<<"Variable name: "<<var_name<<"\n";
	  
	  int thisid;
	  if (namevarmap.count(var_name) == 0)
	  {
	    Var var;
	    var.name = var_name;
	    var.id = varid;
	    thisid = varid;
	    ++varid;
	    if (intblock)
	      var.type = INTEGER;
	    namevarmap[var_name] = var;
	  }
	  else
	    thisid = namevarmap[var_name].id;
	  
	  string con_name_1 = line.substr(14,8);
// 	  cout<<"Constraint name for this variable: "<<con_name_1<<"\n";
	  double coeff_1 = atof(line.substr(24,12).c_str());
// 	  cout<<"Coefficient "<<coeff_1<<"\n";
	  
	  if (con_name_1 == obj_name)
	    namevarmap[var_name].obj = coeff_1;
	  else
	    nameconmap[con_name_1].coeffs.push_back(pair<int,double>(thisid,coeff_1));

	  line.erase(0,39);
	  if (line.size()>11)
	  {
	    string con_name_2 = line.substr(0,8);
// 	    cout<<"Constraint name for this variable: "<<con_name_2<<"\n";
	    double coeff_2 = atof((line.substr(10,12)).c_str());
// 	    cout<<"Coefficient "<<coeff_2<<"\n";
	    
	    if (con_name_2 == obj_name)
	      namevarmap[var_name].obj = coeff_2;
	    else
	      nameconmap[con_name_2].coeffs.push_back(pair<int,double>(thisid,coeff_2));
	  }
	  
	  break;
	}
	case MPS_RHS:
	{
 
	  string con_name_1 = line.substr(14,8);
	  if (con_name_1.length() < 8)
	    con_name_1.append(8-con_name_1.length(),' ');	  
// 	  cout<<"Constraint name: "<<con_name_1<<"\n";
	  double coeff_1 = atof((line.substr(24,12)).c_str());
// 	  cout<<"Coefficient "<<coeff_1<<"\n";

	  //ignore constant offset for objective
	  if (con_name_1 != obj_name)
	    nameconmap[con_name_1].rhs = coeff_1;
	  
	  line.erase(0,39);
	  if (line.size()>11)
	  {
	    string con_name_2 = line.substr(0,8);
	    if (con_name_2.length() < 8)
	      con_name_2.append(8-con_name_2.length(),' ');
// 	    cout<<"Constraint name: "<<con_name_2<<"\n";
	    double coeff_2 = atof((line.substr(10,12)).c_str());
// 	    cout<<"Coefficient "<<coeff_2<<"\n";
	    if (con_name_2 != obj_name)
	      nameconmap[con_name_2].rhs = coeff_2;
	  }

	  break;
	}
	case MPS_RANGES:
	{
	  string con_name_1 = line.substr(14,8);
	  if (con_name_1.length() < 8)
	    con_name_1.append(8-con_name_1.length(),' ');
	  double coeff_1 = atof((line.substr(24,12)).c_str());
	  Con rangecon1 = nameconmap[con_name_1];
	  rangecon1.id = conid;
	  ++conid;
	  rangecon1.name.append("_RANGES");
	  if (rangecon1.sen == GEQ)
	  {
	    rangecon1.sen = LEQ;
	    rangecon1.rhs += abs(coeff_1);
	  }
	  else if (rangecon1.sen == LEQ)
	  {
	    rangecon1.sen = GEQ;
	    rangecon1.rhs -= abs(coeff_1);
	  }
	  else if (rangecon1.sen == EQ)
	    rangecon1.rhs += coeff_1;
	  
	  nameconmap[rangecon1.name] = rangecon1; 
	  
	  line.erase(0,39);
	  if (line.size()>11)
	  {
	    string con_name_2 = line.substr(0,8);
	    if (con_name_2.length() < 8)
	      con_name_2.append(8-con_name_2.length(),' ');
	    double coeff_2 = atof((line.substr(10,12)).c_str());
	    Con rangecon2 = nameconmap[con_name_2];
	    rangecon2.id = conid;
	    ++conid;
	    rangecon2.name.append("_RANGES");
	    if (rangecon2.sen == GEQ)
	    {
	      rangecon2.sen = LEQ;
	      rangecon2.rhs += abs(coeff_2);
	    }
	    else if (rangecon2.sen == LEQ)
	    {
	      rangecon2.sen = GEQ;
	      rangecon2.rhs -= abs(coeff_2);
	    }
	    else if (rangecon2.sen == EQ)
	      rangecon2.rhs += coeff_2;
	    
	    nameconmap[rangecon2.name] = rangecon2;
	    
	  } 
	}
	break;
	case MPS_BOUNDS:
	{
	  string boundtype = line.substr(1,2);
	  string varname = line.substr(14,8);
	  if (varname.length() < 8)
	    varname.append(8-varname.length(),' ');
	  
	  double coeff = 0;
	  if (line.size() > 24)
	    coeff = atof((line.substr(24,12)).c_str());

	  if (boundtype == "LO")
	  {
	    namevarmap[varname].lb = coeff;
	  }
	  else if (boundtype == "UP")
	  {
	    namevarmap[varname].ub = coeff;
	    if (coeff < 0)
	      namevarmap[varname].lb = ROPIMINF;
	  }
	  else if (boundtype == "FX")
	  {
	    namevarmap[varname].lb = coeff;
	    namevarmap[varname].ub = coeff;
	  }
	  else if (boundtype == "FR")
	  {
	    namevarmap[varname].lb = ROPIMINF;
	    namevarmap[varname].ub = ROPIPINF;
	  }
	  else if (boundtype == "MI")
	  {
	    namevarmap[varname].lb = ROPIMINF;
	  }
	  else if (boundtype == "PL")
	  {
	    namevarmap[varname].ub = ROPIPINF;
	  }
	  else if (boundtype == "BV")
	  {
	    namevarmap[varname].lb = 0;
	    namevarmap[varname].ub = 1;
	    namevarmap[varname].type = BINARY;
	  }
	  else if (boundtype == "LI")
	  {
	    namevarmap[varname].lb = coeff;
	    namevarmap[varname].type = INTEGER;
	  }
	  else if (boundtype == "UI")
	  {
	    namevarmap[varname].ub = coeff;
	    namevarmap[varname].type = INTEGER;
	  }
	  else if (boundtype == "SC")
	  {
	    stringstream estring;
	    estring << "In read_mpsfile: Semi-continuous variables currently not supported.\n";
	    estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	    throw Error(estring.str());
	  }
	  else
	  {
	    stringstream estring;
	    estring << "In read_mpsfile: Unknown variable type "<<boundtype<<".\n";
	    estring << "In file "<<file<<", line "<<linenumber;
	    throw Error(estring.str());
	  }
	}
	  break;
	case MPS_NONE:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Line has no header.\n";
	  estring << "In file "<<file<<", line "<<linenumber;
	  throw Error(estring.str());
	  break;
	}
	default:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Unknown error reading mps.\n";
	  estring << "In file "<<file<<", line "<<linenumber;
	  throw Error(estring.str());
	}
      }
    }
  }
  
  for (map<string,Var>::iterator it=namevarmap.begin(); it!=namevarmap.end(); ++it)
    add_var(it->second);
  
  for (map<string,Con>::iterator it=nameconmap.begin(); it!=nameconmap.end(); ++it)
    add_con(it->second);
  
//   cout<<"Reading in done.\n";
  
}

void LPimpl::write_plain(ostream &out)
{
  out<<"***************************\n";
  out<<"LP Name: "<<name<<"\n";
  if (obj_sen == MAX)
    out<<"Max ";
  else
    out<<"Min ";
  if (obj.size() == 0)
    out<<"0";
  else
    for (list<Var*>::iterator it=obj.begin(); it!=obj.end(); ++it)
      out<<(*it)->obj<<"*"<<(*it)->name<<" ";
  out<<"\n";
    
  out<<"s.t.\n";
  
  for (list<Con>::iterator it=cons.begin(); it!=cons.end(); ++it)
  {
    out<<it->name<<": ";
   for (list<pair<int,double> >::iterator it2=it->coeffs.begin(); it2!=it->coeffs.end(); ++it2)
     out<<it2->second<<"*"<<idvarmap[it2->first]->name<<" ";
   if (it->sen == LEQ)
     out<<"<= ";
   else if (it->sen == EQ)
     out<<"= ";
   else
     out<<">= ";
   out<<it->rhs<<"\n";
  }
  
  out<<"Bounds:\n";
  for (list<Var>::iterator it=vars.begin(); it!=vars.end(); ++it)
  {
    if (it->lb == ROPIMINF)
      out<<"-INF";
    else
      out<<it->lb;
    
    out<<" <= "<<it->name<<" <= ";
    
    if (it->ub == ROPIPINF)
      out<<"INF\n";
    else
      out<<it->ub<<"\n";
  }   

  out<<"***************************\n";

}

vector<string> LPimpl::tokenize(string line)
{
  vector<string> retval;
  bool blank = true;
  while (line.size() > 0)
  {
    char front = line[0];
    line.erase(0,1);
    if (blank && front == ' ')
      continue;
    else if (front == ' ')
      blank = true;
    else if (blank)
    {
      retval.push_back(string(1,front));
      blank = false;
    }
    else
      (retval.back()).append(1,front);
  }
  return retval;
};

void LPimpl::read_mpsfile_free(string file)
{
  cout<<"Using free mps.\n";
  
   ifstream in(file.c_str());
  
  MPS_READMODE readmode = MPS_NONE;
  string line;
  int linenumber = 0;
  int varid = 0;
  int conid = 0;
  string obj_name;
  map<string,Var> namevarmap;
  map<string,Con> nameconmap;
  obj_sen = MIN;
  bool intblock = false;
  
  while (!in.eof())
  {
    getline(in,line);
    ++linenumber;
//     cout<<linenumber<<"\n"<<flush;
    
    vector<string> tokens = tokenize(line);
    
    //empty lines
    if (tokens.size() == 0)
      continue;
    
    //comments
    if ((line.c_str())[0]=='*')
      continue;
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTORG'")!=string::npos)
    {
      intblock = true;
      continue;
    }
    
    if (line.find("'MARKER'")!=string::npos && line.find("'INTEND'")!=string::npos)
   {
      intblock = false;
      continue;
    }
    
    //header
    if ((line.c_str())[0]!=' ')
    {
      if (tokens[0] == "NAME")
      {
	if (tokens.size() == 1)
	{
	  stringstream estring;
	  estring << "In read_mpsfile: NAME is not followed by string.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
	}
	name = tokens[1];
	readmode = MPS_NAME;
      }
      else if (tokens[0] == "ROWS")
	readmode = MPS_ROWS;
      else if (tokens[0] == "COLUMNS")
	readmode = MPS_COLUMNS;
      else if (tokens[0] == "RHS")
	readmode = MPS_RHS;
      else if (tokens[0] == "RANGES")
	readmode = MPS_RANGES;
      else if (tokens[0] == "BOUNDS")
	readmode = MPS_BOUNDS;
      else if (tokens[0] == "ENDATA")
	readmode = MPS_NONE;
      else
      {
	  stringstream estring;
	  estring << "In read_mpsfile: Unknown header.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
      }
    }
    else //no header
    {
      switch (readmode)
      {
	case MPS_NAME:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Name is not part of name header.\n";
	  estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	  throw Error(estring.str());
	  break;
	}
	case MPS_ROWS:
	{
	  string conname = tokens[1];

	  switch (tokens[0][0])
	  {
	    case 'N':
	      obj_name = conname;
// 	      cout<<"Name of Objective: "<<obj_name<<"\n";
	      break;
	    case 'L':
	    {
	      Con con;
	      con.sen = LEQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    case 'G':
	    {
	      Con con;
	      con.sen = GEQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    case 'E':
	    {
	      Con con;
	      con.sen = EQ;
	      con.rhs = 0;
	      con.name = conname;
	      con.id = conid;
	      ++conid;
	      nameconmap[conname] = con;
	      break;
	    }
	    default:
	    {
	      stringstream estring;
	      estring << "In read_mpsfile: Unknown row type.\n";
	      estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	      throw Error(estring.str());
	    }
	  }
	}
	  break;
	case MPS_COLUMNS:
	{
	  string var_name = tokens[0];
// 	  cout<<"Variable name: "<<var_name<<"\n";
	  
	  int thisid;
	  if (namevarmap.count(var_name) == 0)
	  {
	    Var var;
	    var.name = var_name;
	    var.id = varid;
	    thisid = varid;
	    ++varid;
	    if (intblock)
	      var.type = INTEGER;
	    namevarmap[var_name] = var;
	  }
	  else
	    thisid = namevarmap[var_name].id;
	  
	  string con_name_1 = tokens[1];
// 	  cout<<"Constraint name for this variable: "<<con_name_1<<"\n";
	  double coeff_1 = atof(tokens[2].c_str());
// 	  cout<<"Coefficient "<<coeff_1<<"\n";
	  
	  if (con_name_1 == obj_name)
	    namevarmap[var_name].obj = coeff_1;
	  else
	    nameconmap[con_name_1].coeffs.push_back(pair<int,double>(thisid,coeff_1));

	  if (tokens.size() > 3)
	  {
	    string con_name_2 = tokens[3];
// 	    cout<<"Constraint name for this variable: "<<con_name_2<<"\n";
	    double coeff_2 = atof(tokens[4].c_str());
// 	    cout<<"Coefficient "<<coeff_2<<"\n";
	    
	    if (con_name_2 == obj_name)
	      namevarmap[var_name].obj = coeff_2;
	    else
	      nameconmap[con_name_2].coeffs.push_back(pair<int,double>(thisid,coeff_2));
	  }
	  
	  break;
	}
	case MPS_RHS:
	{
 
	  string con_name_1 = tokens[1];
// 	  cout<<"Constraint name: "<<con_name_1<<"\n";
	  double coeff_1 = atof(tokens[2].c_str());
// 	  cout<<"Coefficient "<<coeff_1<<"\n";

	  //ignore constant offset for objective
	  if (con_name_1 != obj_name)
	    nameconmap[con_name_1].rhs = coeff_1;
	  
	  if (tokens.size() > 3)
	  {
	    string con_name_2 = tokens[3];
// 	    cout<<"Constraint name: "<<con_name_2<<"\n";
	    double coeff_2 = atof(tokens[4].c_str());
// 	    cout<<"Coefficient "<<coeff_2<<"\n";
	    if (con_name_2 != obj_name)
	      nameconmap[con_name_2].rhs = coeff_2;
	  }

	  break;
	}
	case MPS_RANGES:
	{
	  string con_name_1 = tokens[1];
	  double coeff_1 = atof(tokens[2].c_str());
	  Con rangecon1 = nameconmap[con_name_1];
	  rangecon1.id = conid;
	  ++conid;
	  rangecon1.name.append("_RANGES");
	  if (rangecon1.sen == GEQ)
	  {
	    rangecon1.sen = LEQ;
	    rangecon1.rhs += abs(coeff_1);
	  }
	  else if (rangecon1.sen == LEQ)
	  {
	    rangecon1.sen = GEQ;
	    rangecon1.rhs -= abs(coeff_1);
	  }
	  else if (rangecon1.sen == EQ)
	    rangecon1.rhs += coeff_1;
	  
	  nameconmap[rangecon1.name] = rangecon1; 
	  
	  if (tokens.size() > 3)
	  {
	    string con_name_2 = tokens[3];
	    double coeff_2 = atof(tokens[4].c_str());
	    Con rangecon2 = nameconmap[con_name_2];
	    rangecon2.id = conid;
	    ++conid;
	    rangecon2.name.append("_RANGES");
	    if (rangecon2.sen == GEQ)
	    {
	      rangecon2.sen = LEQ;
	      rangecon2.rhs += abs(coeff_2);
	    }
	    else if (rangecon2.sen == LEQ)
	    {
	      rangecon2.sen = GEQ;
	      rangecon2.rhs -= abs(coeff_2);
	    }
	    else if (rangecon2.sen == EQ)
	      rangecon2.rhs += coeff_2;
	    
	    nameconmap[rangecon2.name] = rangecon2;
	    
	  } 
	}
	break;
	case MPS_BOUNDS:
	{
	  string boundtype = tokens[0];
	  string varname = tokens[2];
	  
	  double coeff = 0;
	  if (tokens.size() > 3)
	    coeff = atof(tokens[3].c_str());

	  if (boundtype == "LO")
	  {
	    namevarmap[varname].lb = coeff;
	  }
	  else if (boundtype == "UP")
	  {
	    namevarmap[varname].ub = coeff;
	    if (coeff < 0)
	      namevarmap[varname].lb = ROPIMINF;
	  }
	  else if (boundtype == "FX")
	  {
	    namevarmap[varname].lb = coeff;
	    namevarmap[varname].ub = coeff;
	  }
	  else if (boundtype == "FR")
	  {
	    namevarmap[varname].lb = ROPIMINF;
	    namevarmap[varname].ub = ROPIPINF;
	  }
	  else if (boundtype == "MI")
	  {
	    namevarmap[varname].lb = ROPIMINF;
	  }
	  else if (boundtype == "PL")
	  {
	    namevarmap[varname].ub = ROPIPINF;
	  }
	  else if (boundtype == "BV")
	  {
	    namevarmap[varname].lb = 0;
	    namevarmap[varname].ub = 1;
	    namevarmap[varname].type = BINARY;
	  }
	  else if (boundtype == "LI")
	  {
	    namevarmap[varname].lb = coeff;
	    namevarmap[varname].type = INTEGER;
	  }
	  else if (boundtype == "UI")
	  {
	    namevarmap[varname].ub = coeff;
	    namevarmap[varname].type = INTEGER;
	  }
	  else if (boundtype == "SC")
	  {
	    stringstream estring;
	    estring << "In read_mpsfile: Semi-continuous variables currently not supported.\n";
	    estring << "In file "<<file<<", line "<<linenumber<<": "<<line;
	    throw Error(estring.str());
	  }
	  else
	  {
	    stringstream estring;
	    estring << "In read_mpsfile: Unknown variable type "<<boundtype<<".\n";
	    estring << "In file "<<file<<", line "<<linenumber;
	    throw Error(estring.str());
	  }
	}
	  break;
	case MPS_NONE:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Line has no header.\n";
	  estring << "In file "<<file<<", line "<<linenumber;
	  throw Error(estring.str());
	  break;
	}
	default:
	{
	  stringstream estring;
	  estring << "In read_mpsfile: Unknown error reading mps.\n";
	  estring << "In file "<<file<<", line "<<linenumber;
	  throw Error(estring.str());
	}
      }
    }
  }
  
  for (map<string,Var>::iterator it=namevarmap.begin(); it!=namevarmap.end(); ++it)
    add_var(it->second);
  
  for (map<string,Con>::iterator it=nameconmap.begin(); it!=nameconmap.end(); ++it)
    add_con(it->second); 
}
