// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "rimpl_light.h"
#include "error.h"
#include "solver.h"
#include <sstream>

using namespace std;
using namespace ROPI;

LP Rimpl_Light::generate_robust()
{
  if (options.solvertype == SOL_UNDEFINED)
    throw Error("In generate_robust: Light Robustness preprocessing solver is unspecified.");

  switch (unctype)
  {
    case UNC_FINITE:
			return generate_finite();
    case UNC_UNDEFINED:
      throw Error("In generate_robust: Uncertainty is undefined.");
    default:
      throw Error("In generate_robust: Unkown uncertainty.");
  }

}

Robustness_Method Rimpl_Light::get_type()
{
  return ROB_LIGHT;
}

double Rimpl_Light::presolve()
{
	Solver solver;
	solver.init(lp, options.solvertype);
	solver.solve();
	
	return solver.get_objective();
}

LP Rimpl_Light::generate_finite()
{
	list<Var>* lpvars = lp->get_vars();
    list<Con>* lpcons = lp->get_cons();
    
    int nr_scenarios = finunc.scenarios.size();
    int nr_vars = lpvars->size();
    int nr_cons = lpcons->size();    
    
	LP rc;
	double budget = options.budget * presolve();
	
	vector<Var> rcvars(nr_vars);
	vector<Con> rccons(nr_cons);	
	
	//nominal variables
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
		rcvars[it->id] = *it;
		rcvars[it->id].obj = 0;
		stringstream varname;
		varname<<rcvars[it->id].name;
		varname<<"-NOM";
		rcvars[it->id].name = varname.str();
	}	
	
	vector<Var> gamma;
	int varcounter = nr_vars+1;
	//relaxation variables
	for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
	{
		Var x;
		x.id = varcounter;
		x.obj = 1;
		stringstream varname;
		varname<<"GAMMA-";
		varname<<varcounter-nr_vars;
		x.name = varname.str();
		++varcounter;
		gamma.push_back(x);
	}
	
	int con_counter = 0;
	
	//budget constraint
	Con budget_con;
	budget_con.id = con_counter;
	budget_con.name = "BUDGET";
	budget_con.rhs = budget;
	++con_counter;
	
	if (lp->get_sense() == MIN)
		budget_con.sen = LEQ;
	else
		budget_con.sen = GEQ;
		
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
		budget_con.coeffs.push_back(make_pair<int,double> (it->id, it->obj) );
		
	//nominal feasibility
	vector<Con> feascons;
	for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
	{
		feascons.push_back(*it);
		stringstream conname;
		conname<<feascons.back().name;
		conname<<"-NOMFEAS";
		feascons.back().name = conname.str();
	
		++con_counter;
	}
	
	//constraint relaxation
	vector<Con> relaxation;	
	for (int i=0; i<nr_scenarios; ++i)
	{
		int inner_con = 0;		
		
		for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		{
			
			if (it->sen == LEQ || it->sen == GEQ)
			{
				Con con = *it;
				con.id = con_counter;
				++con_counter;
				
				for (list<pair<int,double> >::iterator coeffit=con.coeffs.begin(); coeffit!=con.coeffs.end(); ++coeffit)
				{
					if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first)) > 0)
						coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first)];
				}	
			            
				if (finunc.scenarios[i].rhs.count(*it) > 0)
					con.rhs = finunc.scenarios[i].rhs[*it];				
				
				//add gamma to constraint
				if (it->sen == LEQ)
					con.coeffs.push_back(make_pair<int,double>(nr_vars+1+inner_con, -1));
				else
					con.coeffs.push_back(make_pair<int,double>(nr_vars+1+inner_con, 1));
				
				++inner_con;
				relaxation.push_back(con);
			}
			else
			{
				Con con1 = *it;
				con1.id = con_counter;
				++con_counter;
				con1.sen = LEQ;
				
				Con con2 = *it;
				con2.id = con_counter;
				++con_counter;
				con2.sen = GEQ;
				
				for (list<pair<int,double> >::iterator coeffit=con1.coeffs.begin(); coeffit!=con1.coeffs.end(); ++coeffit)
				{
					if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first)) > 0)
						coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first)];
				}
				
				for (list<pair<int,double> >::iterator coeffit=con2.coeffs.begin(); coeffit!=con2.coeffs.end(); ++coeffit)
				{
					if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first)) > 0)
						coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first)];
				}					
			            
				if (finunc.scenarios[i].rhs.count(*it) > 0)
					con1.rhs = finunc.scenarios[i].rhs[*it];
					
				if (finunc.scenarios[i].rhs.count(*it) > 0)
					con2.rhs = finunc.scenarios[i].rhs[*it];							
				
				//add gamma to constraint
				con1.coeffs.push_back(make_pair<int,double>(nr_vars+1+inner_con, -1));
				con2.coeffs.push_back(make_pair<int,double>(nr_vars+1+inner_con, 1));
				
				++inner_con;
				relaxation.push_back(con1);
				relaxation.push_back(con2);
			}
			
			
		}
	}
	
	//add to rc
	for (unsigned int i=0; i<rcvars.size(); ++i)
		rc.add_var(rcvars[i]);

	for (unsigned int i=0; i<gamma.size(); ++i)
		rc.add_var(gamma[i]);
		
	rc.add_con(budget_con);
		
	for (unsigned int i=0; i<feascons.size(); ++i)
		rc.add_con(feascons[i]);
		
	for (unsigned int i=0; i<relaxation.size(); ++i)
		rc.add_con(relaxation[i]);
	
	return rc;
}
