// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "rimpl_recopt.h"
#include "error.h"
#include "solver.h"
#include <sstream>

using namespace std;
using namespace ROPI;

LP Rimpl_Recopt::generate_robust()
{
  if (options.norm==NORM_UNDEFINED)
    throw Error("In generate_robust: Norm is undefined.");
  
  if (options.recobj == REC_UNDEFINED)
	throw Error("In generate_robust: Recovery objective is undefined.");

  if (options.recopt_model == RECOPT_UNDEFINED)
	throw Error("In generate_robust: RecOpt model is unspecified.");

  if (options.solvertype == SOL_UNDEFINED)
    throw Error("In generate_robust: RecOpt preprocessing solver is unspecified.");

	if (options.recopt_model == RECOPT_SIMPLE)
	{
	  switch (unctype)
	  {
	    case UNC_FINITE:
			if (options.norm == NORM_L1 && options.recobj == REC_MEDIAN)
				return generate_finite_l1_median_simple();
			else if (options.norm == NORM_L1 && options.recobj == REC_CENTER)
				return generate_finite_l1_center_simple();
			else if (options.norm == NORM_LINFTY && options.recobj == REC_MEDIAN)
				return generate_finite_linfty_median_simple();
			else if (options.norm == NORM_LINFTY && options.recobj == REC_CENTER)
				return generate_finite_linfty_center_simple();
				
	    case UNC_UNDEFINED:
	      throw Error("In generate_robust: Uncertainty is undefined.");
	    default:
	      throw Error("In generate_robust: Unkown uncertainty.");
	  }
	}
	else if (options.recopt_model == RECOPT_EXTENDED)
	{
	  switch (unctype)
	  {
	    case UNC_FINITE:
			if (options.norm == NORM_L1 && options.recobj == REC_MEDIAN)
				return generate_finite_l1_median_extended();
			else if (options.norm == NORM_L1 && options.recobj == REC_CENTER)
				return generate_finite_l1_center_extended();
			else if (options.norm == NORM_LINFTY && options.recobj == REC_MEDIAN)
				return generate_finite_linfty_median_extended();
			else if (options.norm == NORM_LINFTY && options.recobj == REC_CENTER)
				return generate_finite_linfty_center_extended();

				
	    case UNC_UNDEFINED:
	      throw Error("In generate_robust: Uncertainty is undefined.");
	    default:
	      throw Error("In generate_robust: Unkown uncertainty.");
	  }		
	}
	else
		throw Error("In generate_robust: Unkown RecOpt model.");

}

Robustness_Method Rimpl_Recopt::get_type()
{
  return ROB_RECOPT;
}

vector<double> Rimpl_Recopt::presolve()
{
	int nr_scenarios = finunc.scenarios.size();
	vector<double> sols(nr_scenarios+1);
	
	//nominal solution
	{
		Solver solver;
        solver.init(lp, options.solvertype);
		solver.solve();
		sols[0] = solver.get_objective();
	}
	
	//scenario solutions
	for (int k=0; k<nr_scenarios; ++k)
	{
		list<Var>* lpvars = lp->get_vars();
		list<Con>* lpcons = lp->get_cons();
		int nr_vars = lpvars->size();
		int nr_cons = lpcons->size();
		
		LP scenario_lp;
		vector<Var> scvars(nr_vars);
		vector<Con> sccons(nr_cons);
		
		for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
			scvars[it->id] = *it;
		
		int con_counter = 0;
		for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		{
			sccons[con_counter] = *it;
		    for (list<pair<int,double> >::iterator coeffit=sccons[con_counter].coeffs.begin(); coeffit!=sccons[con_counter].coeffs.end(); ++coeffit)
			{
				if (finunc.scenarios[k].coeffs.count(make_pair<Con,int>(*it,coeffit->first)) > 0)
					coeffit->second = finunc.scenarios[k].coeffs[make_pair<Con,int>(*it,coeffit->first)];
			}
      
			if (finunc.scenarios[k].rhs.count(*it) > 0)
				sccons[con_counter].rhs = finunc.scenarios[k].rhs[*it];
			
			++con_counter;
		}
		
		for (unsigned int i=0; i<scvars.size(); ++i)
		    scenario_lp.add_var(scvars[i]);
		    
		for (unsigned int i=0; i<sccons.size(); ++i)
		    scenario_lp.add_con(sccons[i]);
		
		Solver solver;
        solver.init(&scenario_lp, options.solvertype);
		solver.solve();
		sols[k+1] = solver.get_objective();
		
	}
	
	return sols;
}

vector<vector<double> > Rimpl_Recopt::pre_solutions()
{
	int nr_scenarios = finunc.scenarios.size();
	vector<vector<double> > sols(nr_scenarios+1);
	
	//nominal solution
	{
		Solver solver;
        solver.init(lp, options.solvertype);
		solver.solve();
		sols[0] = solver.get_solution();
	}
	
	//scenario solutions
	for (int k=0; k<nr_scenarios; ++k)
	{
		list<Var>* lpvars = lp->get_vars();
		list<Con>* lpcons = lp->get_cons();
		int nr_vars = lpvars->size();
		int nr_cons = lpcons->size();
		
		LP scenario_lp;
		vector<Var> scvars(nr_vars);
		vector<Con> sccons(nr_cons);
		
		for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
			scvars[it->id] = *it;
		
		int con_counter = 0;
		for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		{
			sccons[con_counter] = *it;
		    for (list<pair<int,double> >::iterator coeffit=sccons[con_counter].coeffs.begin(); coeffit!=sccons[con_counter].coeffs.end(); ++coeffit)
			{
				if (finunc.scenarios[k].coeffs.count(make_pair<Con,int>(*it,coeffit->first)) > 0)
					coeffit->second = finunc.scenarios[k].coeffs[make_pair<Con,int>(*it,coeffit->first)];
			}
      
			if (finunc.scenarios[k].rhs.count(*it) > 0)
				sccons[con_counter].rhs = finunc.scenarios[k].rhs[*it];
			
			++con_counter;
		}
		
		for (unsigned int i=0; i<scvars.size(); ++i)
		    scenario_lp.add_var(scvars[i]);
		    
		for (unsigned int i=0; i<sccons.size(); ++i)
		    scenario_lp.add_con(sccons[i]);
		
		Solver solver;
        solver.init(&scenario_lp, options.solvertype);
		solver.solve();
		sols[k+1] = solver.get_solution();
		
	}
	
	return sols;
}

LP Rimpl_Recopt::generate_finite_l1_median_simple()
{
	list<Var>* lpvars = lp->get_vars();
	list<Con>* lpcons = lp->get_cons();
	LP rc;
	
	vector<vector<double> > sols = pre_solutions();
	int nr_scenarios = finunc.scenarios.size();
	int nr_vars = lpvars->size();
	int nr_cons = lpcons->size();
	
	//location variables
	vector<Var> locvars(nr_vars);
	
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
		stringstream namestream;
		namestream<<"ROBVAR-"<<it->id;
		locvars[it->id].name = namestream.str();
		locvars[it->id].id = it->id;
		locvars[it->id].lb = ROPIMINF;
	}
	
	//absolute values
	vector<vector<Var> > absvars(nr_scenarios+1);
	
	for (int i=0; i<nr_scenarios+1; ++i)
	{
		absvars[i].resize(nr_vars);
		for (int j=0; j<nr_vars; ++j)
		{
			stringstream namestream;
			namestream<<"ABSVAR-"<<i<<"-"<<j;
			absvars[i][j].name = namestream.str();
			absvars[i][j].id = nr_vars + i*nr_vars + j;
			absvars[i][j].obj = 1;
			
		}
	}
	
  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,-1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
    
  for (unsigned int i=0; i<absvars.size(); ++i)
    for (unsigned int j=0; j<absvars[i].size(); ++j)
    rc.add_var(absvars[i][j]);
    
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
    
  rc.set_name(lp->get_name() + "-RECOPT");
  
  return rc; 
    
}

LP Rimpl_Recopt::generate_finite_linfty_median_simple()
{
	list<Var>* lpvars = lp->get_vars();
	list<Con>* lpcons = lp->get_cons();
	LP rc;
	
	vector<vector<double> > sols = pre_solutions();
	int nr_scenarios = finunc.scenarios.size();
	int nr_vars = lpvars->size();
	int nr_cons = lpcons->size();
	
	//location variables
	vector<Var> locvars(nr_vars);
	
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
		stringstream namestream;
		namestream<<"ROBVAR-"<<it->id;
		locvars[it->id].name = namestream.str();
		locvars[it->id].id = it->id;
		locvars[it->id].lb = ROPIMINF;
	}
	
	//absolute values
	vector<Var> absvars(nr_scenarios);
	
	for (int i=0; i<nr_scenarios; ++i)
	{
		stringstream namestream;
		namestream<<"ABSVAR-"<<i;
		absvars[i].name = namestream.str();
		absvars[i].id = nr_vars + i;
		absvars[i].obj = 1;
	}
	
  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,-1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
    
  for (unsigned int i=0; i<absvars.size(); ++i)
    rc.add_var(absvars[i]);
    
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
    
  rc.set_name(lp->get_name() + "-RECOPT");
  
  return rc; 	
}


LP Rimpl_Recopt::generate_finite_l1_center_simple()
{
	list<Var>* lpvars = lp->get_vars();
	list<Con>* lpcons = lp->get_cons();
	LP rc;
	
	vector<vector<double> > sols = pre_solutions();
	int nr_scenarios = finunc.scenarios.size();
	int nr_vars = lpvars->size();
	int nr_cons = lpcons->size();
	
	//location variables
	vector<Var> locvars(nr_vars);
	
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
		stringstream namestream;
		namestream<<"ROBVAR-"<<it->id;
		locvars[it->id].name = namestream.str();
		locvars[it->id].id = it->id;
		locvars[it->id].lb = ROPIMINF;
	}
	
	//absolute values
	vector<vector<Var> > absvars(nr_scenarios+1);
	
	for (int i=0; i<nr_scenarios+1; ++i)
	{
		absvars[i].resize(nr_vars);
		for (int j=0; j<nr_vars; ++j)
		{
			stringstream namestream;
			namestream<<"ABSVAR-"<<i<<"-"<<j;
			absvars[i][j].name = namestream.str();
			absvars[i][j].id = nr_vars + i*nr_vars + j;
			absvars[i][j].obj = 0;
			
		}
	}
	
	
	Var objvar;
	objvar.name = "Objective";
	objvar.id = 2*nr_vars + nr_scenarios*nr_vars;
	objvar.obj = 1;
	
  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,-1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  //objective constraints
  vector<Con> objcons(nr_scenarios+1);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    objcons[i].sen = LEQ;
    objcons[i].name = "OBJCON";
    objcons[i].id = (nr_scenarios + 1)*nr_cons + (nr_scenarios+1)*nr_vars*2 + i;
    for (int j=0; j<nr_vars; ++j)
		objcons[i].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
	objcons[i].coeffs.push_back(make_pair<int,double>(objvar.id,-1));
  }
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
    
  for (unsigned int i=0; i<absvars.size(); ++i)
    for (unsigned int j=0; j<absvars[i].size(); ++j)
    rc.add_var(absvars[i][j]);
    
  rc.add_var(objvar);  
    
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
    
  for (unsigned int i=0; i<objcons.size(); ++i)
    rc.add_con(objcons[i]);
    
  rc.set_name(lp->get_name() + "-RECOPT");
  
  return rc; 
    
}


LP Rimpl_Recopt::generate_finite_linfty_center_simple()
{
	list<Var>* lpvars = lp->get_vars();
	list<Con>* lpcons = lp->get_cons();
	LP rc;
	
	vector<vector<double> > sols = pre_solutions();
	int nr_scenarios = finunc.scenarios.size();
	int nr_vars = lpvars->size();
	int nr_cons = lpcons->size();
	
	//location variables
	vector<Var> locvars(nr_vars);
	
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
		stringstream namestream;
		namestream<<"ROBVAR-"<<it->id;
		locvars[it->id].name = namestream.str();
		locvars[it->id].id = it->id;
		locvars[it->id].lb = ROPIMINF;
	}
	
	//absolute values
	vector<Var> absvars(nr_scenarios+1);
	
	for (int i=0; i<nr_scenarios+1; ++i)
	{
		stringstream namestream;
		namestream<<"ABSVAR-"<<i;
		absvars[i].name = namestream.str();
		absvars[i].id = nr_vars + i;
		absvars[i].obj = 0;
	}
	
	
	Var objvar;
	objvar.name = "Objective";
	objvar.id = nr_vars + nr_scenarios + 1;
	objvar.obj = 1;
	
  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,-1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
      loccons[current].rhs = sols[i][j];
    }
  }
  
  //objective constraints
  vector<Con> objcons(nr_scenarios+1);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    objcons[i].sen = LEQ;
    objcons[i].name = "OBJCON";
    objcons[i].id = (nr_scenarios + 1)*nr_cons + (nr_scenarios+1)*nr_vars*2 + i;
	objcons[i].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
	objcons[i].coeffs.push_back(make_pair<int,double>(objvar.id,-1));
  }
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
    
  for (unsigned int i=0; i<absvars.size(); ++i)
    rc.add_var(absvars[i]);
    
  rc.add_var(objvar);  
    
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
    
  for (unsigned int i=0; i<objcons.size(); ++i)
    rc.add_con(objcons[i]);
    
  rc.set_name(lp->get_name() + "-RECOPT");
  
  return rc; 	
}

LP Rimpl_Recopt::generate_finite_l1_median_extended()
{
	
  vector<double> sols = presolve();	
	
  list<Var>* lpvars = lp->get_vars();
  list<Con>* lpcons = lp->get_cons();
  LP rc;
  
  int nr_scenarios = finunc.scenarios.size();
  int nr_vars = lpvars->size();
  int nr_cons = lpcons->size();
  
  //don't forget the nominal case
  vector<Var> rcvars(nr_vars*(nr_scenarios+1));
  vector<Con> rccons(nr_cons*(nr_scenarios+1));
  
  //nominal case variables
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    rcvars[it->id] = *it;
    rcvars[it->id].obj = 0;
    stringstream varname;
    varname<<rcvars[it->id].name;
    varname<<"-SC-NOM";
    rcvars[it->id].name = varname.str();
  }
  
  //scenario variables
  for (int i=0; i<nr_scenarios; ++i)
  {
    
    for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
    {
      rcvars[it->id+(i+1)*nr_vars] = *it;
      rcvars[it->id+(i+1)*nr_vars].id = it->id+(i+1)*nr_vars;
      rcvars[it->id+(i+1)*nr_vars].obj = 0;
      stringstream varname;
      varname<<rcvars[it->id+(i+1)*nr_vars].name;
      varname<<"-SC-"<<i;
      rcvars[it->id+(i+1)*nr_vars].name = varname.str();
    }
  }
  
  //nominal case constraints
  int con_counter = 0;
  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
  {
    rccons[con_counter] = *it;
    stringstream conname;
    conname<<rccons[con_counter].name;
    conname<<"-SC-NOM";
    rccons[con_counter].name = conname.str();
    ++con_counter;
  }
  
  //scenario constraints
  for (int i=0; i<nr_scenarios; ++i)
  {
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      rccons[con_counter] = *it;
      stringstream conname;
      conname<<rccons[con_counter].name;
      conname<<"-SC-"<<i;
      rccons[con_counter].name = conname.str();
      rccons[con_counter].id = it->id + (i+1)*nr_cons;
      //neue coeffs
      for (list<pair<int,double> >::iterator coeffit=rccons[con_counter].coeffs.begin(); coeffit!=rccons[con_counter].coeffs.end(); ++coeffit)
      {
	coeffit->first += (i+1)*nr_vars;
	if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)) > 0)
	  coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)];
      }
      
      if (finunc.scenarios[i].rhs.count(*it) > 0)
	rccons[con_counter].rhs = finunc.scenarios[i].rhs[*it];
      
      ++con_counter;
 
    }
  }
  
  //location variables
  vector<Var> locvars(nr_vars);
  vector<vector<Var> > absvars(nr_scenarios+1);
  
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    stringstream namestream;
    namestream<<"ROBVAR-"<<it->id;
    locvars[it->id].name = namestream.str();
    locvars[it->id].id = nr_vars*(nr_scenarios+1) + it->id;
    locvars[it->id].lb = ROPIMINF;
  }
  
	//nominal feasibility?
	vector<Con> feascons;
	if (options.nominal_feasibility)
	{
		int feascounter = 0;
		  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		  {
		    feascons.push_back(*it);
		    stringstream conname;
		    conname<<feascons[feascounter].name;
		    conname<<"-NOMFEAS";
		    feascons[feascounter].name = conname.str();
		    
		  for (list<pair<int,double> >::iterator coeffit=feascons[feascounter].coeffs.begin(); coeffit!=feascons[feascounter].coeffs.end(); ++coeffit)
			coeffit->first += (nr_scenarios+1)*nr_vars;
		    
		    ++feascounter;
		  }
		  
	  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	  {
	    locvars[it->id].lb = it->lb;
	    locvars[it->id].ub = it->ub;
        locvars[it->id].type = it->type;
	  }
		
	}  
  
  //absolute values
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    absvars[i].resize(nr_vars);
    for (int j=0; j<nr_vars; ++j)
    {
      stringstream namestream;
      namestream<<"ABSVAR-"<<i<<"-"<<j;
      absvars[i][j].name = namestream.str();
      absvars[i][j].id = nr_vars*(nr_scenarios+2) + i*(nr_vars) + j;
      absvars[i][j].obj = 1;
    }
  }

  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,-1));
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
    }
  }
    
  //optimality constraints
  vector<Con> optcons(nr_scenarios+1);  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
			optcons[i].name = "OPTCON";
			optcons[i].id = loccons.back().id + i + 1;
			optcons[i].sen = EQ;
			optcons[i].rhs = sols[i];
							
			for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
				optcons[i].coeffs.push_back(make_pair<int,double>(it->id+ i*nr_vars, it->obj));
  }
  
   
  //add to lp model
  for (unsigned int i=0; i<rcvars.size(); ++i)
    rc.add_var(rcvars[i]);
    
  for (unsigned int i=0; i<rccons.size(); ++i)
    rc.add_con(rccons[i]);
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
  
  for (unsigned int i=0; i<absvars.size(); ++i)
    for (unsigned int j=0; j<absvars[i].size(); ++j)
    rc.add_var(absvars[i][j]);
  
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
 
  for (unsigned int i=0; i<feascons.size(); ++i)
    rc.add_con(feascons[i]);
    
  for (unsigned int i=0; i<optcons.size(); ++i)
	rc.add_con(optcons[i]);
  
  rc.set_name(lp->get_name() + "-RECOPT");
  
  return rc; 
}

LP Rimpl_Recopt::generate_finite_linfty_median_extended()
{
	
	vector<double> sols = presolve();
	
  list<Var>* lpvars = lp->get_vars();
  list<Con>* lpcons = lp->get_cons();
  LP rc;
  
  int nr_scenarios = finunc.scenarios.size();
  int nr_vars = lpvars->size();
  int nr_cons = lpcons->size();
  
  //don't forget the nominal case
  vector<Var> rcvars(nr_vars*(nr_scenarios+1));
  vector<Con> rccons(nr_cons*(nr_scenarios+1));
  
  //nominal case variables
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    rcvars[it->id] = *it;
    rcvars[it->id].obj = 0;
    stringstream varname;
    varname<<rcvars[it->id].name;
    varname<<"-SC-NOM";
    rcvars[it->id].name = varname.str();
  }
  
  //scenario variables
  for (int i=0; i<nr_scenarios; ++i)
  {
    
    for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
    {
      rcvars[it->id+(i+1)*nr_vars] = *it;
      rcvars[it->id+(i+1)*nr_vars].id = it->id+(i+1)*nr_vars;
      rcvars[it->id+(i+1)*nr_vars].obj = 0;
      stringstream varname;
      varname<<rcvars[it->id+(i+1)*nr_vars].name;
      varname<<"-SC-"<<i;
      rcvars[it->id+(i+1)*nr_vars].name = varname.str();
    }
  }
  
  //nominal case constraints
  int con_counter = 0;
  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
  {
    rccons[con_counter] = *it;
    stringstream conname;
    conname<<rccons[con_counter].name;
    conname<<"-SC-NOM";
    rccons[con_counter].name = conname.str();
    ++con_counter;
  }
  
  //scenario constraints
  for (int i=0; i<nr_scenarios; ++i)
  {
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      rccons[con_counter] = *it;
      stringstream conname;
      conname<<rccons[con_counter].name;
      conname<<"-SC-"<<i;
      rccons[con_counter].name = conname.str();
      rccons[con_counter].id = it->id + (i+1)*nr_cons;
      //neue coeffs
      for (list<pair<int,double> >::iterator coeffit=rccons[con_counter].coeffs.begin(); coeffit!=rccons[con_counter].coeffs.end(); ++coeffit)
      {
	coeffit->first += (i+1)*nr_vars;
	if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)) > 0)
	  coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)];
      }
      
      if (finunc.scenarios[i].rhs.count(*it) > 0)
	rccons[con_counter].rhs = finunc.scenarios[i].rhs[*it];
      
      ++con_counter;
 
    }
  }
  
  //location variables
  vector<Var> locvars(nr_vars);
  vector<Var> absvars(nr_scenarios+1);
  Var objvar;
  
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    stringstream namestream;
    namestream<<"ROBVAR-"<<it->id;
    locvars[it->id].name = namestream.str();
    locvars[it->id].id = nr_vars*(nr_scenarios+1) + it->id;
    locvars[it->id].lb = ROPIMINF;
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
	  stringstream namestream;
	  namestream<<"ABSVAR-"<<i;
	  absvars[i].name = namestream.str();
	  absvars[i].id = nr_vars*(nr_scenarios+2) + i;
	  absvars[i].obj = 1;
  }
  

	//nominal feasibility?
	vector<Con> feascons;
	if (options.nominal_feasibility)
	{
		int feascounter = 0;
		  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		  {
		    feascons.push_back(*it);
		    stringstream conname;
		    conname<<feascons[feascounter].name;
		    conname<<"-NOMFEAS";
		    feascons[feascounter].name = conname.str();
		    
		  for (list<pair<int,double> >::iterator coeffit=feascons[feascounter].coeffs.begin(); coeffit!=feascons[feascounter].coeffs.end(); ++coeffit)
			coeffit->first += (nr_scenarios+1)*nr_vars;
		    
		    ++feascounter;
		  }
		  
	  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	  {
	    locvars[it->id].lb = it->lb;
	    locvars[it->id].ub = it->ub;
        locvars[it->id].type = it->type;
	  }
		
	}
		

  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  vector<Con> objcons(nr_scenarios+1);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,-1));
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
    }
  }
  
  //optimality constraints
  vector<Con> optcons(nr_scenarios+1);  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
			optcons[i].name = "OPTCON";
			optcons[i].id = loccons.back().id + i + 1;
			optcons[i].sen = EQ;
			optcons[i].rhs = sols[i];
							
			for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
				optcons[i].coeffs.push_back(make_pair<int,double>(it->id+ i*nr_vars, it->obj));
  }
    
  //add to lp model
  for (unsigned int i=0; i<rcvars.size(); ++i)
    rc.add_var(rcvars[i]);
    
  for (unsigned int i=0; i<rccons.size(); ++i)
    rc.add_con(rccons[i]);
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
  
  for (unsigned int i=0; i<absvars.size(); ++i)
    rc.add_var(absvars[i]);
  
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
    
  for (unsigned int i=0; i<feascons.size(); ++i)
    rc.add_con(feascons[i]);
  
  for (unsigned int i=0; i<optcons.size(); ++i)
	rc.add_con(optcons[i]);  
  
  rc.set_name(lp->get_name() + "-RECFEAS");
  
  return rc; 
}


LP Rimpl_Recopt::generate_finite_l1_center_extended()
{
	vector<double> sols = presolve();		
	
  list<Var>* lpvars = lp->get_vars();
  list<Con>* lpcons = lp->get_cons();
  LP rc;
  
  int nr_scenarios = finunc.scenarios.size();
  int nr_vars = lpvars->size();
  int nr_cons = lpcons->size();
  
  //don't forget the nominal case
  vector<Var> rcvars(nr_vars*(nr_scenarios+1));
  vector<Con> rccons(nr_cons*(nr_scenarios+1));
  
  //nominal case variables
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    rcvars[it->id] = *it;
    rcvars[it->id].obj = 0;
    stringstream varname;
    varname<<rcvars[it->id].name;
    varname<<"-SC-NOM";
    rcvars[it->id].name = varname.str();
  }
  
  //scenario variables
  for (int i=0; i<nr_scenarios; ++i)
  {
    
    for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
    {
      rcvars[it->id+(i+1)*nr_vars] = *it;
      rcvars[it->id+(i+1)*nr_vars].id = it->id+(i+1)*nr_vars;
      rcvars[it->id+(i+1)*nr_vars].obj = 0;
      stringstream varname;
      varname<<rcvars[it->id+(i+1)*nr_vars].name;
      varname<<"-SC-"<<i;
      rcvars[it->id+(i+1)*nr_vars].name = varname.str();
    }
  }
  
  //nominal case constraints
  int con_counter = 0;
  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
  {
    rccons[con_counter] = *it;
    stringstream conname;
    conname<<rccons[con_counter].name;
    conname<<"-SC-NOM";
    rccons[con_counter].name = conname.str();
    ++con_counter;
  }
  
  //scenario constraints
  for (int i=0; i<nr_scenarios; ++i)
  {
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      rccons[con_counter] = *it;
      stringstream conname;
      conname<<rccons[con_counter].name;
      conname<<"-SC-"<<i;
      rccons[con_counter].name = conname.str();
      rccons[con_counter].id = it->id + (i+1)*nr_cons;
      //neue coeffs
      for (list<pair<int,double> >::iterator coeffit=rccons[con_counter].coeffs.begin(); coeffit!=rccons[con_counter].coeffs.end(); ++coeffit)
      {
	coeffit->first += (i+1)*nr_vars;
	if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)) > 0)
	  coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)];
      }
      
      if (finunc.scenarios[i].rhs.count(*it) > 0)
	rccons[con_counter].rhs = finunc.scenarios[i].rhs[*it];
      
      ++con_counter;
 
    }
  }
  
  //location variables
  vector<Var> locvars(nr_vars);
  vector<vector<Var> > absvars(nr_scenarios+1);
  Var objvar;
  
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    stringstream namestream;
    namestream<<"ROBVAR-"<<it->id;
    locvars[it->id].name = namestream.str();
    locvars[it->id].id = nr_vars*(nr_scenarios+1) + it->id;
    locvars[it->id].lb = ROPIMINF;
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    absvars[i].resize(nr_vars);
    for (int j=0; j<nr_vars; ++j)
    {
      stringstream namestream;
      namestream<<"ABSVAR-"<<i<<"-"<<j;
      absvars[i][j].name = namestream.str();
      absvars[i][j].id = nr_vars*(nr_scenarios+2) + i*(nr_vars) + j;
    }
  }
  
  objvar.name = "ObjVar";
  objvar.id = (2*nr_scenarios + 3)*nr_vars;


	//nominal feasibility?
	vector<Con> feascons;
	if (options.nominal_feasibility)
	{
		int feascounter = 0;
		  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		  {
		    feascons.push_back(*it);
		    stringstream conname;
		    conname<<feascons[feascounter].name;
		    conname<<"-NOMFEAS";
		    feascons[feascounter].name = conname.str();
		    
		  for (list<pair<int,double> >::iterator coeffit=feascons[feascounter].coeffs.begin(); coeffit!=feascons[feascounter].coeffs.end(); ++coeffit)
			coeffit->first += (nr_scenarios+1)*nr_vars;
		    
		    ++feascounter;
		  }
		  
	  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	  {
	    locvars[it->id].lb = it->lb;
	    locvars[it->id].ub = it->ub;
        locvars[it->id].type = it->type;
	  }
		
	}
		

  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  vector<Con> objcons(nr_scenarios+1);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,-1));
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
    }
  }
  
  //center function
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    objcons[i].sen = LEQ;
    objcons[i].name = "OBJCON";
    objcons[i].id = (nr_scenarios + 1)*nr_cons + (nr_scenarios+1)*nr_vars*2 + i;
    for (int j=0; j<nr_vars; ++j)
		objcons[i].coeffs.push_back(make_pair<int,double>(absvars[i][j].id,1));
    objcons[i].coeffs.push_back(make_pair<int,double>(objvar.id,-1));
  }

  objvar.obj = 1;
  
  //optimality constraints
  vector<Con> optcons(nr_scenarios+1);  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
			optcons[i].name = "OPTCON";
			optcons[i].id = objcons.back().id + i + 1;
			optcons[i].sen = EQ;
			optcons[i].rhs = sols[i];
							
			for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
				optcons[i].coeffs.push_back(make_pair<int,double>(it->id+ i*nr_vars, it->obj));
  }  
    
  //add to lp model
  for (unsigned int i=0; i<rcvars.size(); ++i)
    rc.add_var(rcvars[i]);
    
  for (unsigned int i=0; i<rccons.size(); ++i)
    rc.add_con(rccons[i]);
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
  
  for (unsigned int i=0; i<absvars.size(); ++i)
    for (unsigned int j=0; j<absvars[i].size(); ++j)
    rc.add_var(absvars[i][j]);
  
  rc.add_var(objvar);
  
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
  
  for (unsigned int i=0; i<objcons.size(); ++i)
    rc.add_con(objcons[i]);
    
  for (unsigned int i=0; i<feascons.size(); ++i)
    rc.add_con(feascons[i]);

  for (unsigned int i=0; i<optcons.size(); ++i)
	rc.add_con(optcons[i]);  
  
  rc.set_name(lp->get_name() + "-RECFEAS");
  
  return rc; 
}


LP Rimpl_Recopt::generate_finite_linfty_center_extended()
{
		vector<double> sols = presolve();	
	
  list<Var>* lpvars = lp->get_vars();
  list<Con>* lpcons = lp->get_cons();
  LP rc;
  
  int nr_scenarios = finunc.scenarios.size();
  int nr_vars = lpvars->size();
  int nr_cons = lpcons->size();
  
  //don't forget the nominal case
  vector<Var> rcvars(nr_vars*(nr_scenarios+1));
  vector<Con> rccons(nr_cons*(nr_scenarios+1));
  
  //nominal case variables
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    rcvars[it->id] = *it;
    rcvars[it->id].obj = 0;
    stringstream varname;
    varname<<rcvars[it->id].name;
    varname<<"-SC-NOM";
    rcvars[it->id].name = varname.str();
  }
  
  //scenario variables
  for (int i=0; i<nr_scenarios; ++i)
  {
    
    for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
    {
      rcvars[it->id+(i+1)*nr_vars] = *it;
      rcvars[it->id+(i+1)*nr_vars].id = it->id+(i+1)*nr_vars;
      rcvars[it->id+(i+1)*nr_vars].obj = 0;
      stringstream varname;
      varname<<rcvars[it->id+(i+1)*nr_vars].name;
      varname<<"-SC-"<<i;
      rcvars[it->id+(i+1)*nr_vars].name = varname.str();
    }
  }
  
  //nominal case constraints
  int con_counter = 0;
  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
  {
    rccons[con_counter] = *it;
    stringstream conname;
    conname<<rccons[con_counter].name;
    conname<<"-SC-NOM";
    rccons[con_counter].name = conname.str();
    ++con_counter;
  }
  
  //scenario constraints
  for (int i=0; i<nr_scenarios; ++i)
  {
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      rccons[con_counter] = *it;
      stringstream conname;
      conname<<rccons[con_counter].name;
      conname<<"-SC-"<<i;
      rccons[con_counter].name = conname.str();
      rccons[con_counter].id = it->id + (i+1)*nr_cons;
      //neue coeffs
      for (list<pair<int,double> >::iterator coeffit=rccons[con_counter].coeffs.begin(); coeffit!=rccons[con_counter].coeffs.end(); ++coeffit)
      {
	coeffit->first += (i+1)*nr_vars;
	if (finunc.scenarios[i].coeffs.count(make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)) > 0)
	  coeffit->second = finunc.scenarios[i].coeffs[make_pair<Con,int>(*it,coeffit->first-(i+1)*nr_vars)];
      }
      
      if (finunc.scenarios[i].rhs.count(*it) > 0)
	rccons[con_counter].rhs = finunc.scenarios[i].rhs[*it];
      
      ++con_counter;
 
    }
  }
  
  //location variables
  vector<Var> locvars(nr_vars);
  vector<Var> absvars(nr_scenarios+1);
  Var objvar;
  
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
    stringstream namestream;
    namestream<<"ROBVAR-"<<it->id;
    locvars[it->id].name = namestream.str();
    locvars[it->id].id = nr_vars*(nr_scenarios+1) + it->id;
    locvars[it->id].lb = ROPIMINF;
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
	  stringstream namestream;
	  namestream<<"ABSVAR-"<<i;
	  absvars[i].name = namestream.str();
	  absvars[i].id = nr_vars*(nr_scenarios+2) + i;
  }
  
  objvar.name = "ObjVar";
  objvar.id = (nr_scenarios + 2)*nr_vars + nr_scenarios + 1;


	//nominal feasibility?
	vector<Con> feascons;
	if (options.nominal_feasibility)
	{
		int feascounter = 0;
		  for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
		  {
		    feascons.push_back(*it);
		    stringstream conname;
		    conname<<feascons[feascounter].name;
		    conname<<"-NOMFEAS";
		    feascons[feascounter].name = conname.str();
		    
		  for (list<pair<int,double> >::iterator coeffit=feascons[feascounter].coeffs.begin(); coeffit!=feascons[feascounter].coeffs.end(); ++coeffit)
			coeffit->first += (nr_scenarios+1)*nr_vars;
		    
		    ++feascounter;
		  }
		  
	  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	  {
	    locvars[it->id].lb = it->lb;
	    locvars[it->id].ub = it->ub;
        locvars[it->id].type = it->type;
	  }
		
	}
		

  //location constraints
  vector<Con> loccons(2*(nr_scenarios+1)*nr_vars);
  vector<Con> objcons(nr_scenarios+1);
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = LEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,-1));
    }
  }
  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    for (int j=0; j<nr_vars; ++j)
    {
      int current = i*nr_vars + j + (nr_scenarios+1)*nr_vars;
      loccons[current].id = nr_cons*(nr_scenarios+1) + current;
      loccons[current].name = "ADCON";
      loccons[current].sen = GEQ;
      loccons[current].coeffs.push_back(make_pair<int,double>(locvars[j].id,1));
      loccons[current].coeffs.push_back(make_pair<int,double>(rcvars[j].id + i*nr_vars,-1));
      loccons[current].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
    }
  }
  
  //center function
  for (int i=0; i<nr_scenarios+1; ++i)
  {
    objcons[i].sen = LEQ;
    objcons[i].name = "OBJCON";
    objcons[i].id = (nr_scenarios + 1)*nr_cons + (nr_scenarios+1)*nr_vars*2 + i;
	objcons[i].coeffs.push_back(make_pair<int,double>(absvars[i].id,1));
    objcons[i].coeffs.push_back(make_pair<int,double>(objvar.id,-1));
  }

  objvar.obj = 1;
  
  //optimality constraints
  vector<Con> optcons(nr_scenarios+1);  
  for (int i=0; i<nr_scenarios+1; ++i)
  {
			optcons[i].name = "OPTCON";
			optcons[i].id = objcons.back().id + i + 1;
			optcons[i].sen = EQ;
			optcons[i].rhs = sols[i];
							
			for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
				optcons[i].coeffs.push_back(make_pair<int,double>(it->id+ i*nr_vars, it->obj));
  }  
    
  //add to lp model
  for (unsigned int i=0; i<rcvars.size(); ++i)
    rc.add_var(rcvars[i]);
    
  for (unsigned int i=0; i<rccons.size(); ++i)
    rc.add_con(rccons[i]);
  
  for (unsigned int i=0; i<locvars.size(); ++i)
    rc.add_var(locvars[i]);
  
  for (unsigned int i=0; i<absvars.size(); ++i)
    rc.add_var(absvars[i]);
  
  rc.add_var(objvar);
  
  for (unsigned int i=0; i<loccons.size(); ++i)
    rc.add_con(loccons[i]);
  
  for (unsigned int i=0; i<objcons.size(); ++i)
    rc.add_con(objcons[i]);
    
  for (unsigned int i=0; i<feascons.size(); ++i)
    rc.add_con(feascons[i]);
  
  for (unsigned int i=0; i<optcons.size(); ++i)
    rc.add_con(optcons[i]);
  
  rc.set_name(lp->get_name() + "-RECFEAS");
  
  return rc; 
}
