// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "solver.h"
#include "solverimpl.h"
#include "error.h"
#include "solver_gurobi.h"
#include "solver_cplex.h"
#include "solver_xpress.h"

using namespace std;
using namespace ROPI;

Solver::Solver()
{
 solver_impl = 0;
}

Solver::~Solver()
{
  if (solver_impl != 0)
    switch (type)
    {
      case SOL_GUROBI:
	delete (static_cast<Solver_gurobi*>(solver_impl));
	break;
      case SOL_CPLEX:
	delete (static_cast<Solver_cplex*>(solver_impl));
	break;
      case SOL_XPRESS:
	delete (static_cast<Solver_xpress*>(solver_impl));
	break;
      default:
	throw Error("In init: Solver type unknown.");
    }
}

list<SOLVERTYPE> Solver::get_available()
{
 list<SOLVERTYPE> retval;
 if (Solver_gurobi::available())
   retval.push_back(SOL_GUROBI);
 if (Solver_cplex::available())
   retval.push_back(SOL_CPLEX);
 if (Solver_xpress::available())
   retval.push_back(SOL_XPRESS);
 
 return retval;
}

SOLVERSTATUS Solver::get_status()
{
  return solver_impl->get_status();
}

double Solver::get_objective()
{
  if (solver_impl == 0)
    throw Error("In get_objective: Solver is not initialized.");
  return solver_impl->get_objective();
}

void Solver::init(LP* _lp, SOLVERTYPE _type)
{
  type = _type;
 switch (type)
 {
   case SOL_GUROBI: 
     if (Solver_gurobi::available())
       solver_impl = new Solver_gurobi(_lp);
     else
       throw Error("In init: Gurobi solver is not available.");
     break;
   case SOL_CPLEX:
     if (Solver_cplex::available())
       solver_impl = new Solver_cplex(_lp);
     else
       throw Error("In init: Cplex solver is not available.");
     break;
   case SOL_XPRESS:
     if (Solver_xpress::available())
       solver_impl = new Solver_xpress(_lp);
     else
       throw Error("In init: Xpess solver is not available.");
     break;
   default:
     throw Error("In init: Solver type unknown.");
 }  
  solver_impl->init();
}

void Solver::solve()
{
  if (solver_impl == 0)
    throw Error("In solve: Solver is not initialized.");
  solver_impl->solve();
}

void Solver::set_options(SolverOptions _options)
{
  solver_impl->set_options(_options);
}

void Solver::write_solution(std::ostream &out, double eps)
{
  if (solver_impl == 0)
    throw Error("In write_solution: Solver is not initialized.");
  solver_impl->write_solution(out, eps);
}

vector<double> Solver::get_solution()
{
  if (solver_impl == 0)
    throw Error("In get_solution: Solver is not initialized.");
  return solver_impl->get_solution();
}
