// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "solver_cplex.h"
#include "lp.h"
#include "error.h"
#include "ilcplex/ilocplex.h"
#include<iostream>
#include<sstream>

ILOSTLBEGIN

using namespace std;
using namespace ROPI;

Solver_cplex::Solver_cplex(LP* _lp)
{
  lp = _lp;
  model = 0;
  env = 0;
  cplex = 0;
  status = SOL_EMPTY;
}

Solver_cplex::~Solver_cplex()
{
  if (model != 0)
    delete model;
  if (env != 0)
    delete env;
  if (cplex != 0)
    delete cplex;
}

void Solver_cplex::write_solution(ostream &out, double eps)
{
  if (model == 0)
    throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
  
  out<<"Solution:\n";  
  IloNumArray vals(*env, (IloInt) lp->get_vars());
  for (int i=0; i<lp->get_num_vars(); ++i)
  {
    try
    {
      vals[i]  = cplex->getValue((*vars)[i]);
    }
    catch(...)
    {
      vals[i] = 0;
    }
  }
  
  for (int i=0; i<lp->get_num_vars(); ++i)
    if (vals[i] > eps || vals[i] < -eps)
      out<<(*vars)[i].getName()<<" = "<<vals[i]<<"\n";
//   env->out() << "Values        = " << vals << endl;
}

vector<double> Solver_cplex::get_solution()
{
  if (model == 0)
    throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
    
  vector<double> sol(lp->get_num_vars());
  
  IloNumArray vals(*env, (IloInt) lp->get_vars());
  for (int i=0; i<lp->get_num_vars(); ++i)
  {
    try
    {
      vals[i]  = cplex->getValue((*vars)[i]);
    }
    catch(...)
    {
      vals[i] = 0;
    }
  }
  
  for (int i=0; i<lp->get_num_vars(); ++i)
    sol[i] = vals[i];
  
  return sol;
}

double Solver_cplex::get_objective()
{  
  if (model == 0)
   throw Error("In get_objective: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n"; 
  
  return cplex->getObjValue();
}

void Solver_cplex::init()
{
  env = new IloEnv();
  
  try{
    //Model
//     cout<<"Creating CPLEX model.\n"<<flush;
    model = new IloModel(*env);
    
    IloExpr obj(*env);
//     if (lp->get_sense() == MIN)
//       obj = IloAdd(*model, IloMinimize(*env));
//     else
//       obj = IloAdd(*model, IloMaximize(*env));
    
    //Variables
//     cout<<"Creating variables.\n"<<flush;
    list<Var>* lpvars = lp->get_vars();
    vars = new IloNumVarArray(*env);
    int var_counter = 0;
    for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
    {
      IloNumVar::Type type;
      switch (it->type)
      {
	      case CONTINUOUS: type = ILOFLOAT; break;
	      case INTEGER: type = ILOINT; break;
	      case BINARY: type = ILOBOOL; break;
	      default: type = ILOFLOAT;
      }
      
      if (it->lb ==  ROPIMINF && it->ub == ROPIPINF)
	vars->add(IloNumVar(*env, -IloInfinity, IloInfinity, type, (it->name).c_str()));
      else if (it->lb == ROPIMINF)
	vars->add(IloNumVar(*env, -IloInfinity, it->ub, type, (it->name).c_str()));
      else if (it->ub == ROPIPINF)
	vars->add(IloNumVar(*env, it->lb, IloInfinity, type, (it->name).c_str()));
      else
	vars->add(IloNumVar(*env, it->lb, it->ub, type, (it->name).c_str()));
      
      idposmap[it->id] = var_counter;
//       obj.setLinearCoef((*vars)[var_counter], it->obj);
      obj += it->obj * (*vars)[var_counter];
      ++var_counter;
    }
    
    if (lp->get_sense() == MIN)
      model->add(IloMinimize(*env, obj));
    else
      model->add(IloMaximize(*env, obj));
    obj.end();
    
    //Constraints
//     cout<<"Creating constraints.\n"<<flush;
//     IloRangeArray cons(*env);
    list<Con>* lpcons = lp->get_cons();
//     int con_counter = 0;
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      IloExpr expr(*env);
      for (list<pair<int,double> >::iterator it2=it->coeffs.begin(); it2!=it->coeffs.end(); ++it2)
	expr +=  it2->second * (*vars)[idposmap[it2->first]];
      switch (it->sen)
      {
	case LEQ: model->add(expr <= it->rhs); break;
	case EQ: model->add(expr == it->rhs); break;
	case GEQ: model->add(expr >= it->rhs); break;
	default: model->add(expr <= it->rhs);
      }
      expr.end();
// 	    switch (it->sen)
// 	    {
// 	      case LEQ: cons.add(IloRange(*env, -IloInfinity, it->rhs)); break;
// 	      case EQ: cons.add(IloRange(*env, it->rhs, it->rhs)); break;
// 	      case GEQ: cons.add(IloRange(*env, it->rhs, IloInfinity)); break;
// 	      default: cons.add(IloRange(*env, -IloInfinity, it->rhs));
// 	    }
// 	    
// 	    for (list<pair<int,double> >::iterator it2=it->coeffs.begin(); it2!=it->coeffs.end(); ++it2)
// 	    {
// 	      cons[con_counter].setLinearCoef((*vars)[it2->first], it2->second);
//       
// 	    }
//       ++con_counter;
    }
   
//    model->add(cons);
   cplex = new IloCplex(*model);
//    cplex->exportModel("cplexmodel.lp");
   
//    cout<<"Model created successfully.\n"<<flush;
   status = SOL_LOADED;
  }
  catch (IloException& e) {
    stringstream estring;
    estring << "Concert error.\n";
    estring << e << endl;
    throw Error(estring.str());
  }
  catch (...) {
//     cerr << "Unknown exception caught" << endl;
    throw Error("Unknown concert error.");
  }
  //env->end();
  
}

void Solver_cplex::solve()
{
  if (model == 0)
    throw Error("In solve: Model is not initialized.");
  try
  {
    if (options.verbose == false)
      cplex->setOut(env->getNullStream());
    
    if (options.timelimit > 0)
      cplex->setParam(IloCplex::TiLim, options.timelimit);
    
    if (options.threads > 0)
      cplex->setParam(IloCplex::Threads, options.threads);
    
    cplex->solve();
    
    switch (cplex->getCplexStatus())
    {
      case IloCplex::Optimal:
      {
	status = SOL_OPTIMAL;
	//remove comment!
	IloNumArray vals(*env, (IloInt) lp->get_vars());
	list<Var>* lpvars = lp->get_vars();
	
	for (int i=0; i<lp->get_num_vars(); ++i)
	{
	  try
	  {
	    vals[i]  = cplex->getValue((*vars)[i]);
	  }
	  catch(...)
	  {
	    vals[i] = 0;
	  }
	}
// 	cplex->getValues(vals, *vars);
// 	
	for (list<Var>::iterator it = lpvars->begin(); it!=lpvars->end(); ++it)
	  it->solution = vals[idposmap[it->id]]; 
	break;
      }
      case IloCplex::Infeasible: status = SOL_INFEASIBLE; break;
      case IloCplex::Unbounded: status = SOL_UNBOUNDED; break;
      case IloCplex::AbortTimeLim: status = SOL_TIMELIMIT; break;
      default: status = SOL_UNKNOWN;
    }
    
  }
  catch (IloException& e) {
    stringstream estring;
    estring << "Concert error.\n";
    estring << e << endl;
    throw Error(estring.str());
  }
  catch (...) {
//     cerr << "Unknown exception caught" << endl;
    throw Error("Unknown concert error.");
  }
}

bool Solver_cplex::available()
{
  return true;
}
