// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "solver_gurobi.h"
#include "gurobi_c++.h"
#include "lp.h"
#include "error.h"
#include<map>
#include<iostream>
#include<sstream>

using namespace std;
using namespace ROPI;

Solver_gurobi::Solver_gurobi(LP* _lp)
{
  lp = _lp;
  env = 0;
  model = 0;
  status = SOL_EMPTY;
}

Solver_gurobi::~Solver_gurobi()
{
  if (model != 0)
    delete model;
  if (env != 0)
    delete env;
}

void Solver_gurobi::write_solution(ostream &out, double eps)
{
  if (model == 0)
   throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
  
  out<<"Solution:\n";
  
  GRBVar* vars = model->getVars();
  for (int i=0; i<model->get(GRB_IntAttr_NumVars); ++i)
    if (vars[i].get(GRB_DoubleAttr_X) > eps || vars[i].get(GRB_DoubleAttr_X) < -eps)
      out<<vars[i].get(GRB_StringAttr_VarName)<<" = "<<vars[i].get(GRB_DoubleAttr_X)<<"\n";
  delete[] vars;
}

vector<double> Solver_gurobi::get_solution()
{
  if (model == 0)
    throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
    
  vector<double> sol(lp->get_num_vars());
  
  GRBVar* vars = model->getVars();
  for (int i=0; i<model->get(GRB_IntAttr_NumVars); ++i)
    sol[i] = vars[i].get(GRB_DoubleAttr_X);
  delete[] vars;
  
  return sol;
}

double Solver_gurobi::get_objective()
{  
  if (model == 0)
   throw Error("In get_objective: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n"; 
  
  return model->get(GRB_DoubleAttr_ObjVal);
}

void Solver_gurobi::init()
{
  
    try
    {
      
      //Model
//       cout<<"Creating GUROBI model.\n"<<flush;
        env = new GRBEnv();
        model = new GRBModel(*env);
	
	if (lp->get_sense() == MIN)
	  model->set(GRB_IntAttr_ModelSense, 1);
	else
	  model->set(GRB_IntAttr_ModelSense, -1);
	
	model->set(GRB_StringAttr_ModelName, lp->get_name());
	
	//Variables
// 	cout<<"Creating variables.\n"<<flush;
	list<Var>* lpvars = lp->get_vars();
	map<int,GRBVar> varmap;
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
	    char type;
	    switch (it->type)
	    {
	      case CONTINUOUS: type = GRB_CONTINUOUS; break;
	      case INTEGER: type = GRB_INTEGER; break;
	      case BINARY: type = GRB_BINARY; break;
	      default: type = GRB_CONTINUOUS;
	    }
	    
	    if (it->lb == ROPIMINF && it->ub == ROPIPINF)
	      varmap[it->id] = model->addVar(-GRB_INFINITY, GRB_INFINITY, it->obj, type, it->name);
	    else if (it->lb == ROPIMINF)
	      varmap[it->id] = model->addVar(-GRB_INFINITY, it->ub, it->obj, type, it->name);
	    else if (it->ub == ROPIPINF)
	      varmap[it->id] = model->addVar(it->lb, GRB_INFINITY, it->obj, type, it->name);
	    else
	      varmap[it->id] = model->addVar(it->lb, it->ub, it->obj, type, it->name);
// 	    cout<<"Adding variable ("<<it->lb<<","<<it->ub<<","<<it->obj<<","<<type<<","<<it->name<<")\n"<<flush;

	}
	
	model->update();
	
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
	  if (it->start_given)
	      varmap[it->id].set(GRB_DoubleAttr_Start,it->start);
	}
	
	//Constraints
// 	cout<<"Creating constraints.\n"<<flush;
	list<Con>* lpcons = lp->get_cons();
	for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
	{
	    char sense;
	    switch (it->sen)
	    {
	      case LEQ: sense = GRB_LESS_EQUAL; break;
	      case EQ: sense = GRB_EQUAL; break;
	      case GEQ: sense = GRB_GREATER_EQUAL; break;
	      default: sense = GRB_LESS_EQUAL;
	    }
	    
	    GRBLinExpr lhs;
	    for (list<pair<int,double> >::iterator it2=it->coeffs.begin(); it2!=it->coeffs.end(); ++it2)
	    {
// 	      cout<<varmap[it2->first].get(GRB_StringAttr_VarName)<<" "<<flush;
	      lhs += it2->second * varmap[it2->first];
// 	      cout<<lhs<<"\n"<<flush;
	    }
	    
// 	    cout<<sense<<" "<<it->rhs<<"\n"<<flush;
	    model->addConstr(lhs, sense, it->rhs);
	    
	}
	
// 	cout<<"Model created successfully.\n"<<flush;
	status = SOL_LOADED;
    }
    catch(GRBException e)
    {
      stringstream estring;
      estring << "Gurobi error.\n";
      estring << "Error code = " << e.getErrorCode() << endl;
      estring << e.getMessage() << endl;
      throw Error(estring.str());
    }
    catch(...)
    {
//         cout << "Exception during optimization" << endl;
      throw Error("Unknown Gurobi error.");
    }
}

void Solver_gurobi::solve()
{
  if (model == 0)
   throw Error("In solve: Model is not initialized.");
  
  //options
  if (options.verbose == false)
    (model->getEnv()).set(GRB_IntParam_OutputFlag, 0);
    
  if (options.timelimit > 0)
    (model->getEnv()).set(GRB_DoubleParam_TimeLimit, options.timelimit);
  
  if (options.threads > 0)
    (model->getEnv()).set(GRB_IntParam_Threads, options.threads);
  
  if (options.mipfocus > 0)
    (model->getEnv()).set(GRB_IntParam_MIPFocus, options.mipfocus);
  
  if (options.heuristics >=0)
    (model->getEnv()).set(GRB_DoubleParam_Heuristics, options.heuristics);
  
   model->getEnv().set(GRB_IntParam_Method, 3);

  try
  {
model->update();

    model->optimize();
    
    switch (model->get(GRB_IntAttr_Status))
    {
      case GRB_OPTIMAL:
      {
	status = SOL_OPTIMAL;
	GRBVar* vars = model->getVars();
	list<Var>* lpvars = lp->get_vars();
	int count = 0;
	for (list<Var>::iterator it = lpvars->begin(); it!=lpvars->end(); ++it)
	{
	  it->solution = vars[count].get(GRB_DoubleAttr_X);
	  ++count;
	}
	delete[] vars;
	break;
      }
      case GRB_INFEASIBLE: status = SOL_INFEASIBLE; break;
      case GRB_UNBOUNDED: status = SOL_UNBOUNDED; break;
      case GRB_TIME_LIMIT:
	{
		 status = SOL_TIMELIMIT; 
		GRBVar* vars = model->getVars();
		list<Var>* lpvars = lp->get_vars();
		int count = 0;
		for (list<Var>::iterator it = lpvars->begin(); it!= lpvars->end(); ++it)
		{
			it->solution = vars[count].get(GRB_DoubleAttr_X);
			++count;
		}
		delete[] vars;
		break;
	}
      default: status = SOL_UNKNOWN;
    }
    
  }
  catch(GRBException e)
  {
    stringstream estring;
    estring << "Gurobi error.\n";
    estring << "Error code = " << e.getErrorCode() << endl;
    estring << e.getMessage() << endl;
    throw Error(estring.str());
  }
  catch(...)
  {
//         cout << "Exception during optimization" << endl;
    throw Error("Unknown Gurobi error.");
  }

}

bool Solver_gurobi::available()
{
  return true;
}
