// **************************************************
// This file is part of the ROPI Library.
// Written by Marc Goerigk
// Institut für Numerische und Angewandte Mathematik
// Georg-August Universität Göttingen
// **************************************************

#include "solver_xpress.h"
#include "lp.h"
#include "error.h"
#include "xprb_cpp.h"
#include "xprs.h"
#include<map>
#include<iostream>

using namespace std;
using namespace ROPI;
using namespace ::dashoptimization;

Solver_xpress::Solver_xpress(LP* _lp)
{
  lp = _lp;
  model = 0;
  status = SOL_EMPTY;
}

Solver_xpress::~Solver_xpress()
{
  if (model != 0)
    delete model;
}

void Solver_xpress::write_solution(ostream &out, double eps)
{
  if (model == 0)
   throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
  
  out<<"Solution:\n";
  list<Var>* lpvars = lp->get_vars();
  double val;
  	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
	  val = (model->getVarByName((it->name).c_str())).getSol();
	  if (val > eps || val < -eps)
	    out<<it->name<<" = "<<val<<"\n";
	}
  
}

vector<double> Solver_xpress::get_solution()
{
  if (model == 0)
    throw Error("In write_solution: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n";
    
  vector<double> sol(lp->get_num_vars());
  list<Var>* lpvars = lp->get_vars();
  
  int index = 0;
  for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
  {
	sol[index] = (model->getVarByName((it->name).c_str())).getSol();
	++index;
  }
  
  return sol;
}

double Solver_xpress::get_objective()
{  
  if (model == 0)
   throw Error("In get_objective: Model is not initialized.");
  
  if (status != SOL_OPTIMAL)
    cout<<"Warning: Model is not solved to optimality.\n"; 
  
  return model->getObjVal();
}

void Solver_xpress::init()
{
  
//   cout<<"Creating XPRESS model.\n"<<flush;
  model = new XPRBprob((lp->get_name()).c_str());
  if (lp->get_sense() == MIN)
    model->setSense(XPRB_MINIM);
  else
    model->setSense(XPRB_MAXIM);
  
  //Variables
//   cout<<"Creating variables.\n"<<flush;
  list<Var>* lpvars = lp->get_vars();
  map<int,XPRBvar> varmap;
  XPRBexpr obj; //hier noch leer dazu
	for (list<Var>::iterator it=lpvars->begin(); it!=lpvars->end(); ++it)
	{
	  int type;
	  switch (it->type)
	  {
	      case CONTINUOUS: type = XPRB_PL; break;
	      case INTEGER: type = XPRB_UI; break;
	      case BINARY: type = XPRB_BV; break;
	      default: type = XPRB_PL;
	  }
	  
	  if (it->lb == ROPIMINF && it->ub == ROPIPINF)
	    varmap[it->id] = model->newVar((it->name).c_str(), type, -XPRB_INFINITY, XPRB_INFINITY);
	  else if (it->lb == ROPIMINF)
	    varmap[it->id] = model->newVar((it->name).c_str(), type, -XPRB_INFINITY, it->ub);
	  else if (it->ub == ROPIPINF)
	    varmap[it->id] = model->newVar((it->name).c_str(), type, it->lb, XPRB_INFINITY);
	  else
	    varmap[it->id] = model->newVar((it->name).c_str(), type, it->lb, it->ub);
	  
	  obj += (it->obj)*varmap[it->id];
	}
	model->setObj(model->newCtr(obj));
// 	cout<<"OBJECTIVE ";
// 	obj.print();
// 	cout<<"\n";
	
	    //Constraints
//     cout<<"Creating constraints.\n"<<flush;
    list<Con>* lpcons = lp->get_cons();
    for (list<Con>::iterator it=lpcons->begin(); it!=lpcons->end(); ++it)
    {
      XPRBexpr con;
	    
      	    for (list<pair<int,double> >::iterator it2=it->coeffs.begin(); it2!=it->coeffs.end(); ++it2)
	      con+= (it2->second)*varmap[it2->first];
      
     	    switch (it->sen)
	    {
	      case LEQ: model->newCtr(con <= it->rhs); break;
	      case EQ: model->newCtr(con == it->rhs); break;
	      case GEQ: model->newCtr(con >= it->rhs); break;
	      default: model->newCtr(con <= it->rhs);
	    }
    }
    
//    cout<<"Model created successfully.\n"<<flush;
   status = SOL_LOADED;
//    model->print();
//    model->exportProb(XPRB_LP, "achtdamen");
   
}

void Solver_xpress::solve()
{
  if (model == 0)
   throw Error("In solve: Model is not initialized.");
  
  if (options.verbose == false)
    model->setMsgLevel(0);
  
  if (options.timelimit > 0)
    XPRSsetintcontrol(model->getXPRSprob(), XPRS_MAXTIME, options.timelimit);
  
  if (options.threads > 0)
    XPRSsetintcontrol(model->getXPRSprob(), XPRS_THREADS, options.threads);

  //g: MIP
  if (lp->get_lptype() == LPTYPE_LP)
  {
    model->solve("");
    switch (model->getLPStat())
    {
      case XPRB_LP_OPTIMAL:
      {
	status = SOL_OPTIMAL;
	list<Var>* lpvars = lp->get_vars();
	for (list<Var>::iterator it = lpvars->begin(); it!=lpvars->end(); ++it)
	  it->solution = (model->getVarByName((it->name).c_str())).getSol();
	break;
      }
      case XPRB_LP_INFEAS: status = SOL_INFEASIBLE; break;
      case XPRB_LP_UNBOUNDED: status = SOL_UNBOUNDED; break;
      case XPRB_LP_UNFINISHED: status = SOL_TIMELIMIT; break;
      default: status = SOL_UNKNOWN;
    }
  }
  else
  {
    model->solve("g");
    switch (model->getMIPStat())
    {
      case XPRB_MIP_OPTIMAL:
      {
	status = SOL_OPTIMAL;
	list<Var>* lpvars = lp->get_vars();
	for (list<Var>::iterator it = lpvars->begin(); it!=lpvars->end(); ++it)
	  it->solution = (model->getVarByName((it->name).c_str())).getSol();
	break;
      }
      case XPRB_MIP_INFEAS: status = SOL_INFEASIBLE; break;
      case XPRB_MIP_NO_SOL_FOUND: status = SOL_TIMELIMIT; break;
      default: status = SOL_UNKNOWN;
    }
  }

}

bool Solver_xpress::available()
{
  return true;
}
